# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init

import copy
import inspect
import logging
import os
import stat

import pytest

from cloudinit import importer, util
from cloudinit.distros import ubuntu
from cloudinit.event import EventScope, EventType
from cloudinit.helpers import Paths
from cloudinit.sources import (
    EXPERIMENTAL_TEXT,
    METADATA_UNKNOWN,
    REDACT_SENSITIVE_VALUE,
    UNSET,
    DataSource,
    canonical_cloud_id,
    pkl_load,
    redact_sensitive_keys,
)
from cloudinit.user_data import UserDataProcessor
from tests.unittests.helpers import assert_count_equal, mock


class DataSourceTestSubclassNet(DataSource):

    dsname = "MyTestSubclass"
    url_max_wait = 55

    def __init__(
        self,
        sys_cfg,
        distro,
        paths,
        custom_metadata=None,
        custom_userdata=None,
        get_data_retval=True,
    ):
        super(DataSourceTestSubclassNet, self).__init__(sys_cfg, distro, paths)
        self._custom_userdata = custom_userdata
        self._custom_metadata = custom_metadata
        self._get_data_retval = get_data_retval

    def _get_cloud_name(self):
        return "SubclassCloudName"

    def _get_data(self):
        if self._custom_metadata:
            self.metadata = self._custom_metadata
        else:
            self.metadata = {
                "availability_zone": "myaz",
                "local-hostname": "test-subclass-hostname",
                "region": "myregion",
            }
        if self._custom_userdata:
            self.userdata_raw = self._custom_userdata
        else:
            self.userdata_raw = "userdata_raw"
        self.vendordata_raw = "vendordata_raw"
        return self._get_data_retval


class InvalidDataSourceTestSubclassNet(DataSource):
    pass


class TestDataSource:

    @pytest.fixture(autouse=True)
    def fixtures(self, paths):
        self.sys_cfg = {"datasource": {"_undef": {"key1": False}}}
        self.distro = ubuntu.Distro("somedistro", {}, {})
        self.datasource = DataSource(self.sys_cfg, self.distro, paths)

    @pytest.fixture
    def datasource(self, paths):
        return DataSourceTestSubclassNet(self.sys_cfg, self.distro, paths)

    def test_datasource_init(self):
        """DataSource initializes metadata attributes, ds_cfg and ud_proc."""
        assert self.sys_cfg == self.datasource.sys_cfg
        assert self.distro == self.datasource.distro
        assert self.datasource.userdata is None
        assert {} == self.datasource.metadata
        assert self.datasource.userdata_raw is None
        assert self.datasource.vendordata is None
        assert self.datasource.vendordata_raw is None
        assert {"key1": False} == self.datasource.ds_cfg
        assert isinstance(self.datasource.ud_proc, UserDataProcessor)

    def test_datasource_init_gets_ds_cfg_using_dsname(self, paths):
        """Init uses DataSource.dsname for sourcing ds_cfg."""
        sys_cfg = {"datasource": {"MyTestSubclass": {"key2": False}}}
        distro = "distrotest"  # generally should be a Distro object
        datasource = DataSourceTestSubclassNet(sys_cfg, distro, paths)
        assert {"key2": False} == datasource.ds_cfg

    def test_str_is_classname(self, paths):
        """The string representation of the datasource is the classname."""
        assert "DataSource" == str(self.datasource)
        assert "DataSourceTestSubclassNet" == str(
            DataSourceTestSubclassNet("", "", paths)
        )

    def test_datasource_get_url_params_defaults(self):
        """get_url_params default url config settings for the datasource."""
        params = self.datasource.get_url_params()
        assert params.max_wait_seconds == self.datasource.url_max_wait
        assert params.timeout_seconds == self.datasource.url_timeout
        assert params.num_retries == self.datasource.url_retries
        assert (
            params.sec_between_retries
            == self.datasource.url_sec_between_retries
        )

    def test_datasource_get_url_params_subclassed(self, paths):
        """Subclasses can override get_url_params defaults."""
        sys_cfg = {"datasource": {"MyTestSubclass": {"key2": False}}}
        distro = "distrotest"  # generally should be a Distro object
        datasource = DataSourceTestSubclassNet(sys_cfg, distro, paths)
        expected = (
            datasource.url_max_wait,
            datasource.url_timeout,
            datasource.url_retries,
            datasource.url_sec_between_retries,
        )
        url_params = datasource.get_url_params()
        assert self.datasource.get_url_params() != url_params
        assert expected == url_params

    def test_datasource_get_url_params_ds_config_override(self, paths):
        """Datasource configuration options can override url param defaults."""
        sys_cfg = {
            "datasource": {
                "MyTestSubclass": {
                    "max_wait": "1",
                    "timeout": "2",
                    "retries": "3",
                    "sec_between_retries": 4,
                }
            }
        }
        datasource = DataSourceTestSubclassNet(sys_cfg, self.distro, paths)
        expected = (1, 2, 3, 4)
        url_params = datasource.get_url_params()
        assert (
            datasource.url_max_wait,
            datasource.url_timeout,
            datasource.url_retries,
            datasource.url_sec_between_retries,
        ) != url_params
        assert expected == url_params

    def test_datasource_get_url_params_is_zero_or_greater(self, paths):
        """get_url_params ignores timeouts with a value below 0."""
        # Set an override that is below 0 which gets ignored.
        sys_cfg = {"datasource": {"_undef": {"timeout": "-1"}}}
        datasource = DataSource(sys_cfg, self.distro, paths)
        (
            _max_wait,
            timeout,
            _retries,
            _sec_between_retries,
        ) = datasource.get_url_params()
        assert 0 == timeout

    def test_datasource_get_url_uses_defaults_on_errors(self, caplog, paths):
        """On invalid system config values for url_params defaults are used."""
        # All invalid values should be logged
        sys_cfg = {
            "datasource": {
                "_undef": {
                    "max_wait": "nope",
                    "timeout": "bug",
                    "retries": "nonint",
                }
            }
        }
        datasource = DataSource(sys_cfg, self.distro, paths)
        url_params = datasource.get_url_params()
        expected = (
            datasource.url_max_wait,
            datasource.url_timeout,
            datasource.url_retries,
            datasource.url_sec_between_retries,
        )
        assert expected == url_params
        expected_logs = [
            "Config max_wait 'nope' is not an int, using default '-1'",
            "Config timeout 'bug' is not an int, using default '10'",
            "Config retries 'nonint' is not an int, using default '5'",
        ]
        for log in expected_logs:
            assert log in caplog.text

    @mock.patch("cloudinit.distros.net.find_fallback_nic")
    def test_fallback_interface_is_discovered(self, m_get_fallback_nic):
        """The fallback_interface is discovered via find_fallback_nic."""
        m_get_fallback_nic.return_value = "nic9"
        assert "nic9" == self.datasource.distro.fallback_interface

    @mock.patch("cloudinit.sources.net.find_fallback_nic")
    def test_fallback_interface_logs_undiscovered(
        self, m_get_fallback_nic, caplog
    ):
        """Log a warning when fallback_interface can not discover the nic."""
        m_get_fallback_nic.return_value = None  # Couldn't discover nic
        assert self.datasource.distro.fallback_interface is None
        assert (
            mock.ANY,
            logging.WARNING,
            "Did not find a fallback interface on distro: somedistro.",
        ) in caplog.record_tuples

    @mock.patch("cloudinit.sources.net.find_fallback_nic")
    def test_wb_fallback_interface_is_cached(self, m_get_fallback_nic):
        """The fallback_interface is cached and won't be rediscovered."""
        self.datasource.distro.fallback_interface = "nic10"
        assert "nic10" == self.datasource.distro.fallback_interface
        m_get_fallback_nic.assert_not_called()

    def test__get_data_unimplemented(self, paths):
        """Raise an error when _get_data is not implemented."""
        with pytest.raises(
            NotImplementedError,
            match="Subclasses of DataSource must implement _get_data",
        ):
            self.datasource.get_data()
        datasource2 = InvalidDataSourceTestSubclassNet(
            self.sys_cfg, self.distro, paths
        )
        with pytest.raises(
            NotImplementedError,
            match="Subclasses of DataSource must implement _get_data",
        ):
            datasource2.get_data()

    def test_get_data_calls_subclass__get_data(self, datasource):
        """Datasource.get_data uses the subclass' version of _get_data."""
        assert datasource.get_data() is True
        assert {
            "availability_zone": "myaz",
            "local-hostname": "test-subclass-hostname",
            "region": "myregion",
        } == datasource.metadata
        assert "userdata_raw" == datasource.userdata_raw
        assert "vendordata_raw" == datasource.vendordata_raw

    def test_get_hostname_strips_local_hostname_without_domain(
        self, datasource
    ):
        """Datasource.get_hostname strips metadata local-hostname of domain."""
        assert datasource.get_data()
        assert (
            "test-subclass-hostname" == datasource.metadata["local-hostname"]
        )
        assert "test-subclass-hostname" == datasource.get_hostname().hostname
        datasource.metadata["local-hostname"] = "hostname.my.domain.com"
        assert "hostname" == datasource.get_hostname().hostname

    def test_get_hostname_with_fqdn_returns_local_hostname_with_domain(
        self, datasource
    ):
        """Datasource.get_hostname with fqdn set gets qualified hostname."""
        assert datasource.get_data()
        datasource.metadata["local-hostname"] = "hostname.my.domain.com"
        assert (
            "hostname.my.domain.com"
            == datasource.get_hostname(fqdn=True).hostname
        )

    def test_get_hostname_without_metadata_uses_system_hostname(
        self, datasource
    ):
        """Datasource.gethostname runs util.get_hostname when no metadata."""
        assert {} == datasource.metadata
        mock_fqdn = "cloudinit.sources.util.get_fqdn_from_hosts"
        with mock.patch("cloudinit.sources.util.get_hostname") as m_gethost:
            with mock.patch(mock_fqdn) as m_fqdn:
                m_gethost.return_value = "systemhostname.domain.com"
                m_fqdn.return_value = None  # No matching fqdn in /etc/hosts
                assert "systemhostname" == datasource.get_hostname().hostname
                assert (
                    "systemhostname.domain.com"
                    == datasource.get_hostname(fqdn=True).hostname
                )

    def test_get_hostname_without_metadata_returns_none(self, datasource):
        """Datasource.gethostname returns None when metadata_only and no MD."""
        assert {} == datasource.metadata
        mock_fqdn = "cloudinit.sources.util.get_fqdn_from_hosts"
        with mock.patch("cloudinit.sources.util.get_hostname") as m_gethost:
            with mock.patch(mock_fqdn) as m_fqdn:
                assert (
                    datasource.get_hostname(metadata_only=True).hostname
                    is None
                )
                assert (
                    datasource.get_hostname(
                        fqdn=True, metadata_only=True
                    ).hostname
                    is None
                )
        assert [] == m_gethost.call_args_list
        assert [] == m_fqdn.call_args_list

    def test_get_hostname_without_metadata_prefers_etc_hosts(self, datasource):
        """Datasource.gethostname prefers /etc/hosts to util.get_hostname."""
        assert {} == datasource.metadata
        mock_fqdn = "cloudinit.sources.util.get_fqdn_from_hosts"
        with mock.patch("cloudinit.sources.util.get_hostname") as m_gethost:
            with mock.patch(mock_fqdn) as m_fqdn:
                m_gethost.return_value = "systemhostname.domain.com"
                m_fqdn.return_value = "fqdnhostname.domain.com"
                assert "fqdnhostname" == datasource.get_hostname().hostname
                assert (
                    "fqdnhostname.domain.com"
                    == datasource.get_hostname(fqdn=True).hostname
                )

    def test_get_data_does_not_write_instance_data_on_failure(self, paths):
        """get_data does not write INSTANCE_JSON_FILE on get_data False."""
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            paths,
            get_data_retval=False,
        )
        assert datasource.get_data() is False
        json_file = paths.get_runpath("instance_data")
        assert not os.path.exists(
            json_file
        ), f"Found unexpected file {json_file}"

    def test_get_data_writes_json_instance_data_on_success(
        self, datasource, paths
    ):
        """get_data writes INSTANCE_JSON_FILE to run_dir as world readable."""
        sys_info = {
            "python": "3.7",
            "platform": (
                "Linux-5.4.0-24-generic-x86_64-with-Ubuntu-20.04-focal"
            ),
            "uname": [
                "Linux",
                "myhost",
                "5.4.0-24-generic",
                "SMP blah",
                "x86_64",
            ],
            "variant": "ubuntu",
            "dist": ["ubuntu", "20.04", "focal"],
        }
        with mock.patch("cloudinit.util.system_info", return_value=sys_info):
            with mock.patch(
                "cloudinit.sources.canonical_cloud_id",
                return_value="canonical_cloud_id",
            ):
                datasource.get_data()
        json_file = paths.get_runpath("instance_data")
        content = util.load_text_file(json_file)
        expected = {
            "base64_encoded_keys": [],
            "merged_cfg": REDACT_SENSITIVE_VALUE,
            "merged_system_cfg": REDACT_SENSITIVE_VALUE,
            "sensitive_keys": ["merged_cfg", "merged_system_cfg"],
            "sys_info": sys_info,
            "v1": {
                "_beta_keys": ["subplatform"],
                "availability-zone": "myaz",
                "availability_zone": "myaz",
                "cloud_id": "canonical_cloud_id",
                "cloud-name": "subclasscloudname",
                "cloud_name": "subclasscloudname",
                "distro": "ubuntu",
                "distro_release": "focal",
                "distro_version": "20.04",
                "instance-id": "iid-datasource",
                "instance_id": "iid-datasource",
                "local-hostname": "test-subclass-hostname",
                "local_hostname": "test-subclass-hostname",
                "kernel_release": "5.4.0-24-generic",
                "machine": "x86_64",
                "platform": "mytestsubclass",
                "public_ssh_keys": [],
                "python_version": "3.7",
                "region": "myregion",
                "system_platform": (
                    "Linux-5.4.0-24-generic-x86_64-with-Ubuntu-20.04-focal"
                ),
                "subplatform": "unknown",
                "variant": "ubuntu",
            },
            "ds": {
                "_doc": EXPERIMENTAL_TEXT,
                "meta_data": {
                    "availability_zone": "myaz",
                    "local-hostname": "test-subclass-hostname",
                    "region": "myregion",
                },
            },
        }
        assert expected == util.load_json(content)
        file_stat = os.stat(json_file)
        assert 0o644 == stat.S_IMODE(file_stat.st_mode)
        assert expected == util.load_json(content)

    def test_get_data_writes_redacted_public_json_instance_data(self, paths):
        """get_data writes redacted content to public INSTANCE_JSON_FILE."""
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            paths,
            custom_metadata={
                "availability_zone": "myaz",
                "local-hostname": "test-subclass-hostname",
                "region": "myregion",
                "some": {
                    "security-credentials": {
                        "cred1": "sekret",
                        "cred2": "othersekret",
                    }
                },
                "someother": {
                    "nested": {
                        "userData": "HIDE ME",
                    }
                },
                "VENDOR-DAta": "HIDE ME TOO",
            },
        )
        assert_count_equal(
            (
                "combined_cloud_config",
                "merged_cfg",
                "merged_system_cfg",
                "security-credentials",
                "userdata",
                "user-data",
                "user_data",
                "vendordata",
                "vendor-data",
                "ds/vendor_data",
            ),
            datasource.sensitive_metadata_keys,
        )
        sys_info = {
            "python": "3.7",
            "platform": (
                "Linux-5.4.0-24-generic-x86_64-with-Ubuntu-20.04-focal"
            ),
            "uname": [
                "Linux",
                "myhost",
                "5.4.0-24-generic",
                "SMP blah",
                "x86_64",
            ],
            "variant": "ubuntu",
            "dist": ["ubuntu", "20.04", "focal"],
        }
        with mock.patch("cloudinit.util.system_info", return_value=sys_info):
            datasource.get_data()
        json_file = paths.get_runpath("instance_data")
        redacted = util.load_json(util.load_text_file(json_file))
        expected = {
            "base64_encoded_keys": [],
            "merged_cfg": REDACT_SENSITIVE_VALUE,
            "merged_system_cfg": REDACT_SENSITIVE_VALUE,
            "sensitive_keys": [
                "ds/meta_data/VENDOR-DAta",
                "ds/meta_data/some/security-credentials",
                "ds/meta_data/someother/nested/userData",
                "merged_cfg",
                "merged_system_cfg",
            ],
            "sys_info": sys_info,
            "v1": {
                "_beta_keys": ["subplatform"],
                "availability-zone": "myaz",
                "availability_zone": "myaz",
                "cloud-name": "subclasscloudname",
                "cloud_name": "subclasscloudname",
                "cloud_id": "subclasscloudname",
                "distro": "ubuntu",
                "distro_release": "focal",
                "distro_version": "20.04",
                "instance-id": "iid-datasource",
                "instance_id": "iid-datasource",
                "local-hostname": "test-subclass-hostname",
                "local_hostname": "test-subclass-hostname",
                "kernel_release": "5.4.0-24-generic",
                "machine": "x86_64",
                "platform": "mytestsubclass",
                "public_ssh_keys": [],
                "python_version": "3.7",
                "region": "myregion",
                "system_platform": (
                    "Linux-5.4.0-24-generic-x86_64-with-Ubuntu-20.04-focal"
                ),
                "subplatform": "unknown",
                "variant": "ubuntu",
            },
            "ds": {
                "_doc": EXPERIMENTAL_TEXT,
                "meta_data": {
                    "VENDOR-DAta": REDACT_SENSITIVE_VALUE,
                    "availability_zone": "myaz",
                    "local-hostname": "test-subclass-hostname",
                    "region": "myregion",
                    "some": {"security-credentials": REDACT_SENSITIVE_VALUE},
                    "someother": {
                        "nested": {"userData": REDACT_SENSITIVE_VALUE}
                    },
                },
            },
        }
        assert expected == redacted
        file_stat = os.stat(json_file)
        assert 0o644 == stat.S_IMODE(file_stat.st_mode)

    def test_get_data_writes_json_instance_data_sensitive(self, paths):
        """
        get_data writes unmodified data to sensitive file as root-readonly.
        """
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            paths,
            custom_metadata={
                "availability_zone": "myaz",
                "local-hostname": "test-subclass-hostname",
                "region": "myregion",
                "some": {
                    "security-credentials": {
                        "cred1": "sekret",
                        "cred2": "othersekret",
                    }
                },
            },
        )
        sys_info = {
            "python": "3.7",
            "platform": (
                "Linux-5.4.0-24-generic-x86_64-with-Ubuntu-20.04-focal"
            ),
            "uname": [
                "Linux",
                "myhost",
                "5.4.0-24-generic",
                "SMP blah",
                "x86_64",
            ],
            "variant": "ubuntu",
            "dist": ["ubuntu", "20.04", "focal"],
        }

        assert_count_equal(
            (
                "combined_cloud_config",
                "merged_cfg",
                "merged_system_cfg",
                "security-credentials",
                "userdata",
                "user-data",
                "user_data",
                "vendordata",
                "vendor-data",
                "ds/vendor_data",
            ),
            datasource.sensitive_metadata_keys,
        )
        with mock.patch("cloudinit.util.system_info", return_value=sys_info):
            with mock.patch(
                "cloudinit.sources.canonical_cloud_id",
                return_value="canonical-cloud-id",
            ):
                datasource.get_data()
        sensitive_json_file = paths.get_runpath("instance_data_sensitive")
        content = util.load_text_file(sensitive_json_file)
        expected = {
            "base64_encoded_keys": [],
            "merged_cfg": {
                "_doc": (
                    "DEPRECATED: Use merged_system_cfg. Will be dropped "
                    "from 24.1"
                ),
                "datasource": {"_undef": {"key1": False}},
            },
            "merged_system_cfg": {
                "_doc": (
                    "Merged cloud-init system config from "
                    "/etc/cloud/cloud.cfg and /etc/cloud/cloud.cfg.d/"
                ),
                "datasource": {"_undef": {"key1": False}},
            },
            "sensitive_keys": [
                "ds/meta_data/some/security-credentials",
                "merged_cfg",
                "merged_system_cfg",
            ],
            "sys_info": sys_info,
            "v1": {
                "_beta_keys": ["subplatform"],
                "availability-zone": "myaz",
                "availability_zone": "myaz",
                "cloud_id": "canonical-cloud-id",
                "cloud-name": "subclasscloudname",
                "cloud_name": "subclasscloudname",
                "distro": "ubuntu",
                "distro_release": "focal",
                "distro_version": "20.04",
                "instance-id": "iid-datasource",
                "instance_id": "iid-datasource",
                "kernel_release": "5.4.0-24-generic",
                "local-hostname": "test-subclass-hostname",
                "local_hostname": "test-subclass-hostname",
                "machine": "x86_64",
                "platform": "mytestsubclass",
                "public_ssh_keys": [],
                "python_version": "3.7",
                "region": "myregion",
                "subplatform": "unknown",
                "system_platform": (
                    "Linux-5.4.0-24-generic-x86_64-with-Ubuntu-20.04-focal"
                ),
                "variant": "ubuntu",
            },
            "ds": {
                "_doc": EXPERIMENTAL_TEXT,
                "meta_data": {
                    "availability_zone": "myaz",
                    "local-hostname": "test-subclass-hostname",
                    "region": "myregion",
                    "some": {
                        "security-credentials": {
                            "cred1": "sekret",
                            "cred2": "othersekret",
                        }
                    },
                },
            },
        }
        assert_count_equal(expected, util.load_json(content))
        file_stat = os.stat(sensitive_json_file)
        assert 0o600 == stat.S_IMODE(file_stat.st_mode)
        assert expected == util.load_json(content)

    def test_get_data_handles_redacted_unserializable_content(self, paths):
        """get_data warns unserializable content in INSTANCE_JSON_FILE."""
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            paths,
            custom_metadata={"key1": "val1", "key2": {"key2.1": paths}},
        )
        datasource.get_data()
        json_file = paths.get_runpath("instance_data")
        content = util.load_text_file(json_file)
        expected_metadata = {
            "key1": "val1",
            "key2": {
                "key2.1": (
                    "Warning: redacted unserializable type <class"
                    " 'cloudinit.helpers.Paths'>"
                )
            },
        }
        instance_json = util.load_json(content)
        assert expected_metadata == instance_json["ds"]["meta_data"]

    def test_persist_instance_data_writes_ec2_metadata_when_set(
        self, datasource, paths
    ):
        """When ec2_metadata class attribute is set, persist to json."""
        util.ensure_dir(paths.cloud_dir)
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            paths,
        )
        datasource.ec2_metadata = UNSET
        datasource.get_data()
        json_file = paths.get_runpath("instance_data")
        instance_data = util.load_json(util.load_text_file(json_file))
        assert "ec2_metadata" not in instance_data["ds"]
        datasource.ec2_metadata = {"ec2stuff": "is good"}
        datasource.persist_instance_data()
        instance_data = util.load_json(util.load_text_file(json_file))
        assert {"ec2stuff": "is good"} == instance_data["ds"]["ec2_metadata"]

    def test_persist_instance_data_writes_canonical_cloud_id_and_symlink(
        self, tmp_path
    ):
        """canonical-cloud-id class attribute is set, persist to json."""
        tmp = str(tmp_path)
        cloud_dir = os.path.join(tmp, "cloud")
        util.ensure_dir(cloud_dir)
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            Paths({"run_dir": tmp, "cloud_dir": cloud_dir}),
        )
        cloud_id_link = os.path.join(tmp, "cloud-id")
        cloud_id_file = os.path.join(tmp, "cloud-id-my-cloud")
        cloud_id2_file = os.path.join(tmp, "cloud-id-my-cloud2")
        for filename in (cloud_id_file, cloud_id_link, cloud_id2_file):
            assert not os.path.exists(
                filename
            ), f"Unexpected link found {filename}"
        with mock.patch(
            "cloudinit.sources.canonical_cloud_id", return_value="my-cloud"
        ):
            datasource.get_data()
            assert "my-cloud\n" == util.load_text_file(cloud_id_link)
            # A symlink with the generic /run/cloud-init/cloud-id
            # link is present
            assert util.is_link(cloud_id_link)
            datasource.persist_instance_data()
            # cloud-id<cloud-type> not deleted: no cloud-id change
            assert os.path.exists(cloud_id_file)
        # When cloud-id changes, symlink and content change
        with mock.patch(
            "cloudinit.sources.canonical_cloud_id", return_value="my-cloud2"
        ):
            datasource.persist_instance_data()
        assert "my-cloud2\n" == util.load_text_file(cloud_id2_file)
        # Previous cloud-id-<cloud-type> file removed
        assert not os.path.exists(
            cloud_id_file
        ), f"Unexpected {filename} not removed"
        # Generic link persisted which contains canonical-cloud-id as content
        assert util.is_link(cloud_id_link)
        assert "my-cloud2\n" == util.load_text_file(cloud_id_link)

    def test_persist_instance_data_writes_network_json_when_set(
        self, tmp_path
    ):
        """When network_data.json class attribute is set, persist to json."""
        tmp = str(tmp_path)
        cloud_dir = os.path.join(tmp, "cloud")
        util.ensure_dir(cloud_dir)
        paths = Paths({"run_dir": tmp, "cloud_dir": cloud_dir})
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            paths,
        )
        datasource.get_data()
        json_file = paths.get_runpath("instance_data")
        instance_data = util.load_json(util.load_text_file(json_file))
        assert "network_json" not in instance_data["ds"]
        datasource.network_json = {"network_json": "is good"}
        datasource.persist_instance_data()
        instance_data = util.load_json(util.load_text_file(json_file))
        assert {"network_json": "is good"} == instance_data["ds"][
            "network_json"
        ]

    def test_persist_instance_serializes_datasource_pickle(self, tmp_path):
        """obj.pkl is written when instance link present and write_cache."""
        tmp = str(tmp_path)
        cloud_dir = os.path.join(tmp, "cloud")
        util.ensure_dir(cloud_dir)
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            Paths({"run_dir": tmp, "cloud_dir": cloud_dir}),
        )
        pkl_cache_file = os.path.join(cloud_dir, "instance/obj.pkl")
        assert not os.path.exists(pkl_cache_file)
        datasource.network_json = {"network_json": "is good"}
        # No /var/lib/cloud/instance symlink
        datasource.persist_instance_data(write_cache=True)
        assert not os.path.exists(pkl_cache_file)

        # Symlink /var/lib/cloud/instance but write_cache=False
        util.sym_link(cloud_dir, os.path.join(cloud_dir, "instance"))
        datasource.persist_instance_data(write_cache=False)
        assert not os.path.exists(pkl_cache_file)

        # Symlink /var/lib/cloud/instance and write_cache=True
        datasource.persist_instance_data(write_cache=True)
        assert os.path.exists(pkl_cache_file)
        ds = pkl_load(pkl_cache_file)
        assert datasource.network_json == ds.network_json

    def test_get_data_base64encodes_unserializable_bytes(self, tmp_path):
        """On py3, get_data base64encodes any unserializable content."""
        tmp = str(tmp_path)
        paths = Paths({"run_dir": tmp})
        datasource = DataSourceTestSubclassNet(
            self.sys_cfg,
            self.distro,
            paths,
            custom_metadata={"key1": "val1", "key2": {"key2.1": b"\x123"}},
        )
        assert datasource.get_data() is True
        json_file = paths.get_runpath("instance_data")
        content = util.load_text_file(json_file)
        instance_json = util.load_json(content)
        assert_count_equal(
            ["ds/meta_data/key2/key2.1"], instance_json["base64_encoded_keys"]
        )
        assert {"key1": "val1", "key2": {"key2.1": "EjM="}} == instance_json[
            "ds"
        ]["meta_data"]

    def test_get_hostname_subclass_support(self):
        """Validate get_hostname signature on all subclasses of DataSource."""
        base_args = inspect.getfullargspec(DataSource.get_hostname)
        # Import all DataSource subclasses so we can inspect them.
        modules = util.get_modules_from_dir(
            os.path.dirname(os.path.dirname(__file__))
        )
        for _loc, name in modules.items():
            mod_locs, _ = importer.find_module(name, ["cloudinit.sources"], [])
            if mod_locs:
                importer.import_module(mod_locs[0])
        for child in DataSource.__subclasses__():
            if "Test" in child.dsname:
                continue
            assert base_args == inspect.getfullargspec(child.get_hostname), (
                "%s does not implement DataSource.get_hostname params" % child
            )
            for grandchild in child.__subclasses__():
                assert base_args == inspect.getfullargspec(
                    grandchild.get_hostname
                ), (
                    "%s does not implement DataSource.get_hostname params"
                    % grandchild
                )

    def test_clear_cached_attrs_resets_cached_attr_class_attributes(self):
        """Class attributes listed in cached_attr_defaults are reset."""
        count = 0
        # Setup values for all cached class attributes
        for attr, value in self.datasource.cached_attr_defaults:
            setattr(self.datasource, attr, count)
            count += 1
        self.datasource._dirty_cache = True
        self.datasource.clear_cached_attrs()
        for attr, value in self.datasource.cached_attr_defaults:
            assert value == getattr(self.datasource, attr)

    def test_clear_cached_attrs_noops_on_clean_cache(self):
        """Class attributes listed in cached_attr_defaults are reset."""
        count = 0
        # Setup values for all cached class attributes
        for attr, _ in self.datasource.cached_attr_defaults:
            setattr(self.datasource, attr, count)
            count += 1
        self.datasource._dirty_cache = False  # Fake clean cache
        self.datasource.clear_cached_attrs()
        count = 0
        for attr, _ in self.datasource.cached_attr_defaults:
            assert count == getattr(self.datasource, attr)
            count += 1

    def test_clear_cached_attrs_skips_non_attr_class_attributes(self):
        """Skip any cached_attr_defaults which aren't class attributes."""
        self.datasource._dirty_cache = True
        self.datasource.clear_cached_attrs(attr_defaults=(("some", "value"),))
        assert not hasattr(self.datasource, "some")

    def test_clear_cached_attrs_of_custom_attrs(self):
        """Custom attr_values can be passed to clear_cached_attrs."""
        self.datasource._dirty_cache = True
        cached_attr_name = self.datasource.cached_attr_defaults[0][0]
        setattr(self.datasource, cached_attr_name, "himom")
        self.datasource.myattr = "orig"
        self.datasource.clear_cached_attrs(
            attr_defaults=(("myattr", "updated"),)
        )
        assert "himom" == getattr(self.datasource, cached_attr_name)
        assert "updated" == self.datasource.myattr

    @mock.patch.dict(
        DataSource.default_update_events,
        {EventScope.NETWORK: {EventType.BOOT_NEW_INSTANCE}},
    )
    @mock.patch.dict(
        DataSource.supported_update_events,
        {EventScope.NETWORK: {EventType.BOOT_NEW_INSTANCE}},
    )
    def test_update_metadata_only_acts_on_supported_update_events(self):
        """update_metadata_if_supported wont get_data on unsupported events."""
        assert {
            EventScope.NETWORK: set([EventType.BOOT_NEW_INSTANCE])
        } == self.datasource.default_update_events

        fake_get_data = mock.Mock()
        self.datasource.get_data = fake_get_data
        assert not self.datasource.update_metadata_if_supported(
            source_event_types=[EventType.BOOT]
        )
        assert [] == fake_get_data.call_args_list

    @mock.patch.dict(
        DataSource.supported_update_events,
        {EventScope.NETWORK: {EventType.BOOT_NEW_INSTANCE}},
    )
    def test_update_metadata_returns_true_on_supported_update_event(
        self, caplog
    ):
        """update_metadata_if_supported returns get_data on supported events"""

        def fake_get_data():
            return True

        self.datasource.get_data = fake_get_data
        self.datasource._network_config = "something"
        self.datasource._dirty_cache = True
        assert self.datasource.update_metadata_if_supported(
            source_event_types=[
                EventType.BOOT,
                EventType.BOOT_NEW_INSTANCE,
            ]
        )
        assert UNSET == self.datasource._network_config

        assert (
            mock.ANY,
            logging.DEBUG,
            "Update datasource metadata and network config due to"
            " events: boot-new-instance",
        ) in caplog.record_tuples


class TestRedactSensitiveData:
    def test_redact_sensitive_data_noop_when_no_sensitive_keys_present(self):
        """When sensitive_keys is absent or empty from metadata do nothing."""
        md = {"my": "data"}
        assert md == redact_sensitive_keys(md, redact_value="redacted")
        md["sensitive_keys"] = []
        assert md == redact_sensitive_keys(md, redact_value="redacted")

    def test_redact_sensitive_data_redacts_exact_match_name(self):
        """Only exact matched sensitive_keys are redacted from metadata."""
        md = {
            "sensitive_keys": ["md/secure"],
            "md": {"secure": "s3kr1t", "insecure": "publik"},
        }
        secure_md = copy.deepcopy(md)
        secure_md["md"]["secure"] = "redacted"
        assert secure_md == redact_sensitive_keys(md, redact_value="redacted")

    def test_redact_sensitive_data_does_redacts_with_default_string(self):
        """When redact_value is absent, REDACT_SENSITIVE_VALUE is used."""
        md = {
            "sensitive_keys": ["md/secure"],
            "md": {"secure": "s3kr1t", "insecure": "publik"},
        }
        secure_md = copy.deepcopy(md)
        secure_md["md"]["secure"] = "redacted for non-root user"
        assert secure_md == redact_sensitive_keys(md)


class TestCanonicalCloudID:
    def test_cloud_id_returns_platform_on_unknowns(self):
        """When region and cloud_name are unknown, return platform."""
        assert "platform" == canonical_cloud_id(
            cloud_name=METADATA_UNKNOWN,
            region=METADATA_UNKNOWN,
            platform="platform",
        )

    def test_cloud_id_returns_platform_on_none(self):
        """When region and cloud_name are unknown, return platform."""
        assert "platform" == canonical_cloud_id(
            cloud_name=None, region=None, platform="platform"
        )

    def test_cloud_id_returns_cloud_name_on_unknown_region(self):
        """When region is unknown, return cloud_name."""
        for region in (None, METADATA_UNKNOWN):
            assert "cloudname" == canonical_cloud_id(
                cloud_name="cloudname", region=region, platform="platform"
            )

    def test_cloud_id_returns_platform_on_unknown_cloud_name(self):
        """When region is set but cloud_name is unknown return cloud_name."""
        assert "platform" == canonical_cloud_id(
            cloud_name=METADATA_UNKNOWN,
            region="region",
            platform="platform",
        )

    def test_cloud_id_aws_based_on_region_and_cloud_name(self):
        """When cloud_name is aws, return proper cloud-id based on region."""
        assert "aws-china" == canonical_cloud_id(
            cloud_name="aws", region="cn-north-1", platform="platform"
        )
        assert "aws" == canonical_cloud_id(
            cloud_name="aws", region="us-east-1", platform="platform"
        )
        assert "aws-gov" == canonical_cloud_id(
            cloud_name="aws", region="us-gov-1", platform="platform"
        )
        assert "!aws" == canonical_cloud_id(
            cloud_name="!aws", region="us-gov-1", platform="platform"
        )

    def test_cloud_id_azure_based_on_region_and_cloud_name(self):
        """Report cloud-id when cloud_name is azure and region is in china."""
        assert "azure-china" == canonical_cloud_id(
            cloud_name="azure", region="chinaeast", platform="platform"
        )
        assert "azure" == canonical_cloud_id(
            cloud_name="azure", region="!chinaeast", platform="platform"
        )



# Copyright (C) 2016 Canonical Ltd.
#
# Author: Scott Moser <scott.moser@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init

import base64
import os
from collections import OrderedDict
from textwrap import dedent

import pytest

from cloudinit import subp, util
from cloudinit.sources import DataSourceOVF as dsovf
from tests.unittests.helpers import mock

MPATH = "cloudinit.sources.DataSourceOVF."

NOT_FOUND = None

OVF_ENV_CONTENT = """<?xml version="1.0" encoding="UTF-8"?>
<Environment xmlns="http://schemas.dmtf.org/ovf/environment/1"
  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
  xmlns:oe="http://schemas.dmtf.org/ovf/environment/1"
  xsi:schemaLocation="http://schemas.dmtf.org/ovf/environment/1 ../dsp8027.xsd"
  oe:id="WebTier">
  <!-- Information about hypervisor platform -->
  <oe:PlatformSection>
      <Kind>ESX Server</Kind>
      <Version>3.0.1</Version>
      <Vendor>VMware, Inc.</Vendor>
      <Locale>en_US</Locale>
  </oe:PlatformSection>
  <!--- Properties defined for this virtual machine -->
  <PropertySection>
{properties}
  </PropertySection>
</Environment>
"""


def fill_properties(props, template=OVF_ENV_CONTENT):
    lines = []
    prop_tmpl = '<Property oe:key="{key}" oe:value="{val}"/>'
    for key, val in props.items():
        lines.append(prop_tmpl.format(key=key, val=val))
    indent = "        "
    properties = "".join([indent + line + "\n" for line in lines])
    return template.format(properties=properties)


class TestReadOvfEnv:
    def test_with_b64_userdata(self):
        user_data = "#!/bin/sh\necho hello world\n"
        user_data_b64 = base64.b64encode(user_data.encode()).decode()
        props = {
            "user-data": user_data_b64,
            "password": "passw0rd",
            "instance-id": "inst-001",
        }
        env = fill_properties(props)
        md, ud, cfg = dsovf.read_ovf_environment(env)
        assert {"instance-id": "inst-001"} == md
        assert user_data.encode() == ud
        assert {"password": "passw0rd"} == cfg

    def test_with_non_b64_userdata(self):
        user_data = "my-user-data"
        props = {"user-data": user_data, "instance-id": "inst-001"}
        env = fill_properties(props)
        md, ud, cfg = dsovf.read_ovf_environment(env)
        assert {"instance-id": "inst-001"} == md
        assert user_data.encode() == ud
        assert {} == cfg

    def test_with_no_userdata(self):
        props = {"password": "passw0rd", "instance-id": "inst-001"}
        env = fill_properties(props)
        md, ud, cfg = dsovf.read_ovf_environment(env)
        assert {"instance-id": "inst-001"} == md
        assert {"password": "passw0rd"} == cfg
        assert ud is None

    def test_with_b64_network_config_enable_read_network(self):
        network_config = dedent(
            """\
        network:
           version: 2
           ethernets:
              nics:
                 nameservers:
                    addresses:
                    - 127.0.0.53
                    search:
                    - eng.vmware.com
                    - vmware.com
                 match:
                    name: eth*
                 gateway4: 10.10.10.253
                 dhcp4: false
                 addresses:
                 - 10.10.10.1/24
        """
        )
        network_config_b64 = base64.b64encode(network_config.encode()).decode()
        props = {
            "network-config": network_config_b64,
            "password": "passw0rd",
            "instance-id": "inst-001",
        }
        env = fill_properties(props)
        md, ud, cfg = dsovf.read_ovf_environment(env, True)
        assert "inst-001" == md["instance-id"]
        assert {"password": "passw0rd"} == cfg
        assert {
            "version": 2,
            "ethernets": {
                "nics": {
                    "nameservers": {
                        "addresses": ["127.0.0.53"],
                        "search": ["eng.vmware.com", "vmware.com"],
                    },
                    "match": {"name": "eth*"},
                    "gateway4": "10.10.10.253",
                    "dhcp4": False,
                    "addresses": ["10.10.10.1/24"],
                }
            },
        } == md["network-config"]
        assert ud is None

    def test_with_non_b64_network_config_enable_read_network(self):
        network_config = dedent(
            """\
        network:
           version: 2
           ethernets:
              nics:
                 nameservers:
                    addresses:
                    - 127.0.0.53
                    search:
                    - eng.vmware.com
                    - vmware.com
                 match:
                    name: eth*
                 gateway4: 10.10.10.253
                 dhcp4: false
                 addresses:
                 - 10.10.10.1/24
        """
        )
        props = {
            "network-config": network_config,
            "password": "passw0rd",
            "instance-id": "inst-001",
        }
        env = fill_properties(props)
        md, ud, cfg = dsovf.read_ovf_environment(env, True)
        assert {"instance-id": "inst-001"} == md
        assert {"password": "passw0rd"} == cfg
        assert ud is None

    def test_with_b64_network_config_disable_read_network(self):
        network_config = dedent(
            """\
        network:
           version: 2
           ethernets:
              nics:
                 nameservers:
                    addresses:
                    - 127.0.0.53
                    search:
                    - eng.vmware.com
                    - vmware.com
                 match:
                    name: eth*
                 gateway4: 10.10.10.253
                 dhcp4: false
                 addresses:
                 - 10.10.10.1/24
        """
        )
        network_config_b64 = base64.b64encode(network_config.encode()).decode()
        props = {
            "network-config": network_config_b64,
            "password": "passw0rd",
            "instance-id": "inst-001",
        }
        env = fill_properties(props)
        md, ud, cfg = dsovf.read_ovf_environment(env)
        assert {"instance-id": "inst-001"} == md
        assert {"password": "passw0rd"} == cfg
        assert ud is None


class TestDatasourceOVF:
    @pytest.fixture
    def ds(self, paths):
        return dsovf.DataSourceOVF(sys_cfg={}, distro={}, paths=paths)

    def test_get_data_seed_dir(self, ds, paths):
        """Platform info properly reports when getting data from seed dir."""
        # Write ovf-env.xml seed file
        ovf_env = os.path.join(paths.seed_dir, "ovf-env.xml")
        util.write_file(ovf_env, OVF_ENV_CONTENT)

        assert "ovf" == ds.cloud_name
        assert "ovf" == ds.platform_type
        with mock.patch(MPATH + "transport_vmware_guestinfo") as m_guestd:
            with mock.patch(MPATH + "transport_iso9660") as m_iso9660:
                m_iso9660.return_value = NOT_FOUND
                m_guestd.return_value = NOT_FOUND
                assert ds.get_data()
                assert (
                    "ovf (%s/ovf-env.xml)" % paths.seed_dir == ds.subplatform
                )

    @pytest.mark.parametrize(
        ["guestinfo", "iso"],
        [
            pytest.param(False, True, id="vmware_guestinfo"),
            pytest.param(True, False, id="iso9660"),
        ],
    )
    @mock.patch("cloudinit.subp.subp")
    @mock.patch("cloudinit.sources.DataSource.persist_instance_data")
    def test_get_data_vmware_guestinfo_with_network_config(
        self, m_persist, m_subp, guestinfo, iso, ds, paths
    ):
        network_config = dedent(
            """\
        network:
           version: 2
           ethernets:
              nics:
                 nameservers:
                    addresses:
                    - 127.0.0.53
                    search:
                    - vmware.com
                 match:
                    name: eth*
                 gateway4: 10.10.10.253
                 dhcp4: false
                 addresses:
                 - 10.10.10.1/24
        """
        )
        network_config_b64 = base64.b64encode(network_config.encode()).decode()
        props = {
            "network-config": network_config_b64,
            "password": "passw0rd",
            "instance-id": "inst-001",
        }
        env = fill_properties(props)
        with mock.patch(
            MPATH + "transport_vmware_guestinfo",
            return_value=env if guestinfo else NOT_FOUND,
        ):
            with mock.patch(
                MPATH + "transport_iso9660",
                return_value=env if iso else NOT_FOUND,
            ):
                assert ds.get_data() is True
                assert "inst-001" == ds.metadata["instance-id"]
                assert {
                    "version": 2,
                    "ethernets": {
                        "nics": {
                            "nameservers": {
                                "addresses": ["127.0.0.53"],
                                "search": ["vmware.com"],
                            },
                            "match": {"name": "eth*"},
                            "gateway4": "10.10.10.253",
                            "dhcp4": False,
                            "addresses": ["10.10.10.1/24"],
                        }
                    },
                } == ds.network_config


class TestTransportIso9660:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker):
        self.m_find_devs_with = mocker.patch("cloudinit.util.find_devs_with")
        self.m_mounts = mocker.patch("cloudinit.util.mounts")
        self.m_mount_cb = mocker.patch("cloudinit.util.mount_cb")
        self.m_get_ovf_env = mocker.patch(
            "cloudinit.sources.DataSourceOVF.get_ovf_env"
        )
        self.m_get_ovf_env.return_value = ("myfile", "mycontent")

    def test_find_already_mounted(self):
        """Check we call get_ovf_env from on matching mounted devices"""
        mounts = {
            "/dev/sr9": {
                "fstype": "iso9660",
                "mountpoint": "wark/media/sr9",
                "opts": "ro",
            }
        }
        self.m_mounts.return_value = mounts

        assert "mycontent" == dsovf.transport_iso9660()

    def test_find_already_mounted_skips_non_iso9660(self):
        """Check we call get_ovf_env ignoring non iso9660"""
        mounts = {
            "/dev/xvdb": {
                "fstype": "vfat",
                "mountpoint": "wark/foobar",
                "opts": "defaults,noatime",
            },
            "/dev/xvdc": {
                "fstype": "iso9660",
                "mountpoint": "wark/media/sr9",
                "opts": "ro",
            },
        }
        # We use an OrderedDict here to ensure we check xvdb before xvdc
        # as we're not mocking the regex matching, however, if we place
        # an entry in the results then we can be reasonably sure that
        # we're skipping an entry which fails to match.
        self.m_mounts.return_value = OrderedDict(
            sorted(mounts.items(), key=lambda t: t[0])
        )

        assert "mycontent" == dsovf.transport_iso9660()

    def test_find_already_mounted_matches_kname(self):
        """Check we dont regex match on basename of the device"""
        mounts = {
            "/dev/foo/bar/xvdc": {
                "fstype": "iso9660",
                "mountpoint": "wark/media/sr9",
                "opts": "ro",
            }
        }
        # we're skipping an entry which fails to match.
        self.m_mounts.return_value = mounts

        assert NOT_FOUND == dsovf.transport_iso9660()

    def test_mount_cb_called_on_blkdevs_with_iso9660(self):
        """Check we call mount_cb on blockdevs with iso9660 only"""
        self.m_mounts.return_value = {}
        self.m_find_devs_with.return_value = ["/dev/sr0"]
        self.m_mount_cb.return_value = ("myfile", "mycontent")

        assert "mycontent" == dsovf.transport_iso9660()
        self.m_mount_cb.assert_called_with(
            "/dev/sr0", dsovf.get_ovf_env, mtype="iso9660"
        )

    def test_mount_cb_called_on_blkdevs_with_iso9660_check_regex(self):
        """Check we call mount_cb on blockdevs with iso9660 and match regex"""
        self.m_mounts.return_value = {}
        self.m_find_devs_with.return_value = [
            "/dev/abc",
            "/dev/my-cdrom",
            "/dev/sr0",
        ]
        self.m_mount_cb.return_value = ("myfile", "mycontent")

        assert "mycontent" == dsovf.transport_iso9660()
        self.m_mount_cb.assert_called_with(
            "/dev/sr0", dsovf.get_ovf_env, mtype="iso9660"
        )

    def test_mount_cb_not_called_no_matches(self):
        """Check we don't call mount_cb if nothing matches"""
        self.m_mounts.return_value = {}
        self.m_find_devs_with.return_value = ["/dev/vg/myovf"]

        assert NOT_FOUND == dsovf.transport_iso9660()
        assert 0 == self.m_mount_cb.call_count

    def test_mount_cb_called_require_iso_false(self):
        """Check we call mount_cb on blockdevs with require_iso=False"""
        self.m_mounts.return_value = {}
        self.m_find_devs_with.return_value = ["/dev/xvdz"]
        self.m_mount_cb.return_value = ("myfile", "mycontent")

        assert "mycontent" == dsovf.transport_iso9660(require_iso=False)

        self.m_mount_cb.assert_called_with(
            "/dev/xvdz", dsovf.get_ovf_env, mtype=None
        )

    def test_maybe_cdrom_device_none(self):
        """Test maybe_cdrom_device returns False for none/empty input"""
        assert not dsovf.maybe_cdrom_device(None)
        assert not dsovf.maybe_cdrom_device("")

    def test_maybe_cdrom_device_non_string_exception(self):
        """Test maybe_cdrom_device raises ValueError on non-string types"""
        with pytest.raises(ValueError):
            dsovf.maybe_cdrom_device({"a": "eleven"})

    def test_maybe_cdrom_device_false_on_multi_dir_paths(self):
        """Test maybe_cdrom_device is false on /dev[/.*]/* paths"""
        assert not dsovf.maybe_cdrom_device("/dev/foo/sr0")
        assert not dsovf.maybe_cdrom_device("foo/sr0")
        assert not dsovf.maybe_cdrom_device("../foo/sr0")
        assert not dsovf.maybe_cdrom_device("../foo/sr0")

    def test_maybe_cdrom_device_true_on_hd_partitions(self):
        """Test maybe_cdrom_device is false on /dev/hd[a-z][0-9]+ paths"""
        assert dsovf.maybe_cdrom_device("/dev/hda1")
        assert dsovf.maybe_cdrom_device("hdz9")

    def test_maybe_cdrom_device_true_on_valid_relative_paths(self):
        """Test maybe_cdrom_device normalizes paths"""
        assert dsovf.maybe_cdrom_device("/dev/wark/../sr9")
        assert dsovf.maybe_cdrom_device("///sr0")
        assert dsovf.maybe_cdrom_device("/sr0")
        assert dsovf.maybe_cdrom_device("//dev//hda")

    def test_maybe_cdrom_device_true_on_xvd_partitions(self):
        """Test maybe_cdrom_device returns true on xvd*"""
        assert dsovf.maybe_cdrom_device("/dev/xvda")
        assert dsovf.maybe_cdrom_device("/dev/xvda1")
        assert dsovf.maybe_cdrom_device("xvdza1")


@mock.patch(MPATH + "subp.which")
@mock.patch(MPATH + "subp.subp")
class TestTransportVmwareGuestinfo:
    """Test the com.vmware.guestInfo transport implemented in
    transport_vmware_guestinfo."""

    rpctool = "vmware-rpctool"
    rpctool_path = "/not/important/vmware-rpctool"
    vmtoolsd_path = "/not/important/vmtoolsd"

    def test_without_vmware_rpctool_and_without_vmtoolsd_returns_notfound(
        self, m_subp, m_which
    ):
        m_which.side_effect = [None, None]
        assert NOT_FOUND == dsovf.transport_vmware_guestinfo()
        assert (
            0 == m_subp.call_count
        ), "subp should not be called if no rpctool in path."

    def test_without_vmware_rpctool_and_with_vmtoolsd_returns_found(
        self, m_subp, m_which
    ):
        m_which.side_effect = [self.vmtoolsd_path, None]
        m_subp.side_effect = [(fill_properties({}), "")]
        assert NOT_FOUND != dsovf.transport_vmware_guestinfo()
        assert (
            1 == m_subp.call_count
        ), "subp should be called once bc/ rpctool missing."

    def test_notfound_on_exit_code_1(self, m_subp, m_which, caplog):
        """If vmware-rpctool exits 1, then must return not found."""
        m_which.side_effect = [None, self.rpctool_path]
        m_subp.side_effect = subp.ProcessExecutionError(
            stdout="", stderr="No value found", exit_code=1, cmd=["unused"]
        )
        assert NOT_FOUND == dsovf.transport_vmware_guestinfo()
        assert 1 == m_subp.call_count
        assert (
            "WARNING" not in caplog.text
        ), "exit code of 1 by rpctool should not cause warning."

    def test_notfound_if_no_content_but_exit_zero(self, m_subp, m_which):
        """If vmware-rpctool exited 0 with no stdout is normal not-found.

        This isn't actually a case I've seen. normally on "not found",
        rpctool would exit 1 with 'No value found' on stderr.  But cover
        the case where it exited 0 and just wrote nothing to stdout.
        """
        m_which.side_effect = [None, self.rpctool_path]
        m_subp.return_value = ("", "")
        assert NOT_FOUND == dsovf.transport_vmware_guestinfo()
        assert 1 == m_subp.call_count

    def test_notfound_and_warns_on_unexpected_exit_code(
        self, m_subp, m_which, caplog
    ):
        """If vmware-rpctool exits non zero or 1, warnings should be logged."""
        m_which.side_effect = [None, self.rpctool_path]
        m_subp.side_effect = subp.ProcessExecutionError(
            stdout=None, stderr="No value found", exit_code=2, cmd=["unused"]
        )
        assert NOT_FOUND == dsovf.transport_vmware_guestinfo()
        assert 1 == m_subp.call_count
        assert (
            "WARNING" in caplog.text
        ), "exit code of 2 by rpctool should log WARNING."

    def test_found_when_guestinfo_present(self, m_subp, m_which):
        """When there is a ovf info, transport should return it."""
        m_which.side_effect = [None, self.rpctool_path]
        content = fill_properties({})
        m_subp.return_value = (content, "")
        assert content == dsovf.transport_vmware_guestinfo()
        assert 1 == m_subp.call_count

    def test_vmware_rpctool_fails_and_vmtoolsd_fails(
        self, m_subp, m_which, caplog
    ):
        """When vmware-rpctool fails and vmtoolsd fails"""
        m_which.side_effect = [self.vmtoolsd_path, self.rpctool_path]
        m_subp.side_effect = [
            subp.ProcessExecutionError(
                stdout="", stderr="No value found", exit_code=1, cmd=["unused"]
            ),
            subp.ProcessExecutionError(
                stdout="", stderr="No value found", exit_code=1, cmd=["unused"]
            ),
        ]
        assert NOT_FOUND == dsovf.transport_vmware_guestinfo()
        assert 2 == m_subp.call_count
        assert (
            "WARNING" not in caplog.text
        ), "exit code of 1 by rpctool and vmtoolsd should not cause warning."

    def test_vmware_rpctool_fails_and_vmtoolsd_success(self, m_subp, m_which):
        """When vmware-rpctool fails but vmtoolsd succeeds"""
        m_which.side_effect = [self.vmtoolsd_path, self.rpctool_path]
        m_subp.side_effect = [
            subp.ProcessExecutionError(
                stdout="", stderr="No value found", exit_code=1, cmd=["unused"]
            ),
            (fill_properties({}), ""),
        ]
        assert NOT_FOUND != dsovf.transport_vmware_guestinfo()
        assert 2 == m_subp.call_count


#



import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.util import verify_clean_boot, verify_clean_log

USER_DATA = """\
#cloud-config-archive
- type: "text/cloud-boothook"
  content: |
    #!/bin/sh
    echo "this is from a boothook." > /var/tmp/boothook.txt
- type: "text/cloud-config"
  content: |
    bootcmd:
    - echo "this is from a cloud-config." > /var/tmp/bootcmd.txt
"""


@pytest.mark.ci
@pytest.mark.user_data(USER_DATA)
def test_cloud_config_archive(client: IntegrationInstance):
    """Basic correctness test for #cloud-config-archive."""
    log = client.read_from_file("/var/log/cloud-init.log")
    assert "this is from a boothook." in client.read_from_file(
        "/var/tmp/boothook.txt"
    )
    assert "this is from a cloud-config." in client.read_from_file(
        "/var/tmp/bootcmd.txt"
    )
    verify_clean_log(log)
    verify_clean_boot(client)



# Copyright (C) 2015 Canonical Ltd.
# Copyright (C) 2006-2024 Broadcom. All Rights Reserved.
# Broadcom Confidential. The term "Broadcom" refers to Broadcom Inc.
# and/or its subsidiaries.
#
# Author: Sankar Tanguturi <stanguturi@vmware.com>
#         Pengpeng Sun <pengpeng.sun@broadcom.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import logging
import sys
import textwrap

import pytest

from cloudinit.sources.helpers.vmware.imc.boot_proto import BootProtoEnum
from cloudinit.sources.helpers.vmware.imc.config import Config
from cloudinit.sources.helpers.vmware.imc.config_file import (
    ConfigFile as WrappedConfigFile,
)
from cloudinit.sources.helpers.vmware.imc.config_nic import NicConfigurator
from cloudinit.sources.helpers.vmware.imc.guestcust_util import (
    get_network_data_from_vmware_cust_cfg,
    get_non_network_data_from_vmware_cust_cfg,
)
from tests.helpers import cloud_init_project_dir

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
logger = logging.getLogger(__name__)


def ConfigFile(path: str):
    return WrappedConfigFile(cloud_init_project_dir(path))


class TestVmwareConfigFile:
    def test_utility_methods(self):
        """Tests basic utility methods of ConfigFile class"""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        cf.clear()
        assert 0 == len(cf), "clear size"

        cf._insertKey("  PASSWORD|-PASS ", "  foo  ")
        cf._insertKey("BAR", "   ")

        assert 2 == len(cf), "insert size"
        assert "foo" == cf["PASSWORD|-PASS"], "password"
        assert "PASSWORD|-PASS" in cf, "hasPassword"
        assert "FOO" not in cf, "hasFoo"
        assert "BAR" in cf, "hasBar"

    def test_configfile_without_instance_id(self):
        """
        Tests instance id is None when configuration file has no instance id
        """
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)

        md1, _ = get_non_network_data_from_vmware_cust_cfg(conf)
        assert "instance-id" not in md1

        (md2, _) = get_non_network_data_from_vmware_cust_cfg(conf)
        assert "instance-id" not in md2

    def test_configfile_with_instance_id(self):
        """Tests instance id get from configuration file"""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic-instance-id.cfg")
        conf = Config(cf)

        (md1, _) = get_non_network_data_from_vmware_cust_cfg(conf)
        assert md1["instance-id"] == conf.instance_id, "instance-id"

        (md2, _) = get_non_network_data_from_vmware_cust_cfg(conf)
        assert md2["instance-id"] == conf.instance_id, "instance-id"

    def test_configfile_static_2nics(self):
        """Tests Config class for a configuration with two static NICs."""
        cf = ConfigFile("tests/data/vmware/cust-static-2nic.cfg")

        conf = Config(cf)

        assert "myhost1" == conf.host_name, "hostName"
        assert "Africa/Abidjan" == conf.timezone, "tz"

        assert ["10.20.145.1", "10.20.145.2"] == conf.name_servers, "dns"
        assert [
            "eng.vmware.com",
            "proxy.vmware.com",
        ] == conf.dns_suffixes, "suffixes"

        nics = conf.nics
        ipv40 = nics[0].staticIpv4

        assert 2 == len(nics), "nics"
        assert "NIC1" == nics[0].name, "nic0"
        assert "00:50:56:a6:8c:08" == nics[0].mac, "mac0"
        assert BootProtoEnum.STATIC == nics[0].bootProto, "bootproto0"
        assert "10.20.87.154" == ipv40[0].ip, "ipv4Addr0"
        assert "255.255.252.0" == ipv40[0].netmask, "ipv4Mask0"
        assert 2 == len(ipv40[0].gateways), "ipv4Gw0"
        assert "10.20.87.253" == ipv40[0].gateways[0], "ipv4Gw0_0"
        assert "10.20.87.105" == ipv40[0].gateways[1], "ipv4Gw0_1"

        assert 1 == len(nics[0].staticIpv6), "ipv6Cnt0"
        assert "fc00:10:20:87::154" == nics[0].staticIpv6[0].ip, "ipv6Addr0"

        assert "NIC2" == nics[1].name, "nic1"
        assert not nics[1].staticIpv6, "ipv61 dhcp"

    def test_config_file_dhcp_2nics(self):
        """Tests Config class for a configuration with two DHCP NICs."""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        conf = Config(cf)
        nics = conf.nics
        assert 2 == len(nics), "nics"
        assert "NIC1" == nics[0].name, "nic0"
        assert "00:50:56:a6:8c:08" == nics[0].mac, "mac0"
        assert BootProtoEnum.DHCP == nics[0].bootProto, "bootproto0"

    def test_config_password(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        cf._insertKey("PASSWORD|-PASS", "test-password")
        cf._insertKey("PASSWORD|RESET", "no")

        conf = Config(cf)
        assert "test-password" == conf.admin_password, "password"
        assert not conf.reset_password, "do not reset password"

    def test_config_reset_passwd(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        cf._insertKey("PASSWORD|-PASS", "test-password")
        cf._insertKey("PASSWORD|RESET", "random")

        conf = Config(cf)
        with pytest.raises(ValueError):
            assert None is conf.reset_password

        cf.clear()
        cf._insertKey("PASSWORD|RESET", "yes")
        assert 1 == len(cf), "insert size"

        conf = Config(cf)
        assert conf.reset_password, "reset password"

    def test_get_config_nameservers(self):
        """Tests DNS and nameserver settings in a configuration."""
        cf = ConfigFile("tests/data/vmware/cust-static-2nic.cfg")

        config = Config(cf)

        network_config = get_network_data_from_vmware_cust_cfg(config, False)

        assert 2 == network_config.get("version")

        ethernets = network_config.get("ethernets")

        for _, config in ethernets.items():
            assert isinstance(config, dict)
            name_servers = config.get("nameservers").get("addresses")
            dns_suffixes = config.get("nameservers").get("search")
            assert ["10.20.145.1", "10.20.145.2"] == name_servers, "dns"
            assert [
                "eng.vmware.com",
                "proxy.vmware.com",
            ] == dns_suffixes, "suffixes"

    def test_get_config_dns_suffixes(self):
        """Tests if get_network_from_vmware_cust_cfg properly
        generates nameservers and dns settings from a
        specified configuration"""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        config = Config(cf)

        network_config = get_network_data_from_vmware_cust_cfg(config, False)

        assert 2 == network_config.get("version")

        ethernets = network_config.get("ethernets")

        for _, config in ethernets.items():
            assert isinstance(config, dict)
            name_servers = config.get("nameservers").get("addresses")
            dns_suffixes = config.get("nameservers").get("search")
            assert None is name_servers, "dns"
            assert ["eng.vmware.com"] == dns_suffixes, "suffixes"

    def test_get_nics_list_dhcp(self):
        """Tests if NicConfigurator properly calculates ethernets
        for a configuration with a list of DHCP NICs"""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        config = Config(cf)

        nicConfigurator = NicConfigurator(
            config.nics, config.name_servers, config.dns_suffixes, False
        )
        ethernets_dict = nicConfigurator.generate()

        assert isinstance(ethernets_dict, dict)
        assert 2 == len(ethernets_dict), "number of ethernets"

        for name, config in ethernets_dict.items():
            if name == "NIC1":
                assert "00:50:56:a6:8c:08" == config.get("match").get(
                    "macaddress"
                ), "mac address of NIC1"
                assert True is config.get("wakeonlan"), "wakeonlan of NIC1"
                assert True is config.get("dhcp4"), "DHCPv4 enablement of NIC1"
                assert False is config.get("dhcp4-overrides").get(
                    "use-dns"
                ), "use-dns enablement for dhcp4-overrides of NIC1"
            if name == "NIC2":
                assert "00:50:56:a6:5a:de" == config.get("match").get(
                    "macaddress"
                ), "mac address of NIC2"
                assert True is config.get("wakeonlan"), "wakeonlan of NIC2"
                assert True is config.get("dhcp4"), "DHCPv4 enablement of NIC2"
                assert False is config.get("dhcp4-overrides").get(
                    "use-dns"
                ), "use-dns enablement for dhcp4-overrides of NIC2"

    def test_get_nics_list_static(self):
        """Tests if NicConfigurator properly calculates ethernets
        for a configuration with 2 static NICs"""
        cf = ConfigFile("tests/data/vmware/cust-static-2nic.cfg")

        config = Config(cf)

        nicConfigurator = NicConfigurator(
            config.nics, config.name_servers, config.dns_suffixes, False
        )
        ethernets_dict = nicConfigurator.generate()

        assert isinstance(ethernets_dict, dict)
        assert 2 == len(ethernets_dict), "number of ethernets"

        for name, config in ethernets_dict.items():
            print(config)
            if name == "NIC1":
                assert "00:50:56:a6:8c:08" == config.get("match").get(
                    "macaddress"
                ), "mac address of NIC1"
                assert True is config.get("wakeonlan"), "wakeonlan of NIC1"
                assert False is config.get(
                    "dhcp4"
                ), "DHCPv4 enablement of NIC1"
                assert False is config.get(
                    "dhcp6"
                ), "DHCPv6 enablement of NIC1"
                assert [
                    "10.20.87.154/22",
                    "fc00:10:20:87::154/64",
                ] == config.get("addresses"), "IP addresses of NIC1"
                assert [
                    {"to": "10.20.84.0/22", "via": "10.20.87.253"},
                    {"to": "10.20.84.0/22", "via": "10.20.87.105"},
                    {
                        "to": "fc00:10:20:87::/64",
                        "via": "fc00:10:20:87::253",
                    },
                ] == config.get("routes"), "routes of NIC1"
            if name == "NIC2":
                assert "00:50:56:a6:ef:7d" == config.get("match").get(
                    "macaddress"
                ), "mac address of NIC2"
                assert True is config.get("wakeonlan"), "wakeonlan of NIC2"
                assert False is config.get(
                    "dhcp4"
                ), "DHCPv4 enablement of NIC2"
                assert ["192.168.6.102/16"] == config.get(
                    "addresses"
                ), "IP addresses of NIC2"
                assert [
                    {"to": "192.168.0.0/16", "via": "192.168.0.10"},
                ] == config.get("routes"), "routes of NIC2"

    def test_custom_script(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        assert conf.custom_script_name is None
        cf._insertKey("CUSTOM-SCRIPT|SCRIPT-NAME", "test-script")
        conf = Config(cf)
        assert "test-script" == conf.custom_script_name

    def test_post_gc_status(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        assert not conf.post_gc_status
        cf._insertKey("MISC|POST-GC-STATUS", "YES")
        conf = Config(cf)
        assert conf.post_gc_status

    def test_no_default_run_post_script(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        assert not conf.default_run_post_script
        cf._insertKey("MISC|DEFAULT-RUN-POST-CUST-SCRIPT", "NO")
        conf = Config(cf)
        assert not conf.default_run_post_script

    def test_yes_default_run_post_script(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        cf._insertKey("MISC|DEFAULT-RUN-POST-CUST-SCRIPT", "yes")
        conf = Config(cf)
        assert conf.default_run_post_script


class TestVmwareNetConfig:
    """Test conversion of vmware config to cloud-init config."""

    def _get_NicConfigurator(self, tmp_path, text):
        tmp_file = tmp_path / ".config"
        tmp_file.write_text(text)
        cfg = Config(ConfigFile(tmp_file))
        return NicConfigurator(
            cfg.nics,
            cfg.name_servers,
            cfg.dns_suffixes,
            use_system_devices=False,
        )

    def test_static_nic_without_ipv4_netmask(self, tmp_path):
        """netmask is optional for static ipv4 configuration."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = myhost1
            DOMAINNAME = eng.vmware.com

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:a6:8c:08
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 10.20.87.154
            """
        )
        nc = self._get_NicConfigurator(tmp_path, config)
        assert {
            "NIC1": {
                "match": {"macaddress": "00:50:56:a6:8c:08"},
                "wakeonlan": True,
                "dhcp4": False,
                "addresses": ["10.20.87.154/32"],
                "set-name": "NIC1",
            }
        } == nc.generate()

    def test_static_nic_without_ipv6_netmask(self, tmp_path):
        """netmask is mandatory for static ipv6 configuration."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = myhost1
            DOMAINNAME = eng.vmware.com

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:a6:8c:08
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 10.20.87.154
            IPv6ADDR|1 = fc00:10:20:87::154
            """
        )
        nc = self._get_NicConfigurator(tmp_path, config)
        with pytest.raises(ValueError):
            nc.generate()

    def test_non_primary_nic_with_gateway(self, tmp_path):
        """A non primary nic set can have a gateway."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = myhost1
            DOMAINNAME = eng.vmware.com

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:a6:8c:08
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 10.20.87.154
            NETMASK = 255.255.252.0
            GATEWAY = 10.20.87.253
            """
        )
        nc = self._get_NicConfigurator(tmp_path, config)
        assert {
            "NIC1": {
                "match": {"macaddress": "00:50:56:a6:8c:08"},
                "wakeonlan": True,
                "dhcp4": False,
                "addresses": ["10.20.87.154/22"],
                "routes": [{"to": "10.20.84.0/22", "via": "10.20.87.253"}],
                "set-name": "NIC1",
            }
        } == nc.generate()

    def test_cust_non_primary_nic_with_gateway_(self, tmp_path):
        """A customer non primary nic set can have a gateway."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = static-debug-vm
            DOMAINNAME = cluster.local

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:ac:d1:8a
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 100.115.223.75
            NETMASK = 255.255.255.0
            GATEWAY = 100.115.223.254


            [DNS]
            DNSFROMDHCP=no

            NAMESERVER|1 = 8.8.8.8

            [DATETIME]
            UTC = yes
            """
        )
        nc = self._get_NicConfigurator(tmp_path, config)
        assert {
            "NIC1": {
                "match": {"macaddress": "00:50:56:ac:d1:8a"},
                "wakeonlan": True,
                "dhcp4": False,
                "addresses": ["100.115.223.75/24"],
                "routes": [
                    {"to": "100.115.223.0/24", "via": "100.115.223.254"}
                ],
                "set-name": "NIC1",
                "nameservers": {"addresses": ["8.8.8.8"]},
            }
        } == nc.generate()

    def test_a_primary_nic_with_gateway(self, tmp_path):
        """A primary nic set can have a gateway."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = myhost1
            DOMAINNAME = eng.vmware.com

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:a6:8c:08
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 10.20.87.154
            NETMASK = 255.255.252.0
            PRIMARY = true
            GATEWAY = 10.20.87.253
            """
        )
        nc = self._get_NicConfigurator(tmp_path, config)
        assert {
            "NIC1": {
                "match": {"macaddress": "00:50:56:a6:8c:08"},
                "wakeonlan": True,
                "dhcp4": False,
                "addresses": ["10.20.87.154/22"],
                "routes": [{"to": "0.0.0.0/0", "via": "10.20.87.253"}],
                "set-name": "NIC1",
            }
        } == nc.generate()

    def test_meta_data(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        assert conf.meta_data_name is None
        cf._insertKey("CLOUDINIT|METADATA", "test-metadata")
        conf = Config(cf)
        assert "test-metadata" == conf.meta_data_name

    def test_user_data(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        assert conf.user_data_name is None
        cf._insertKey("CLOUDINIT|USERDATA", "test-userdata")
        conf = Config(cf)
        assert "test-userdata" == conf.user_data_name



# This file is part of cloud-init. See LICENSE file for license information.
import pytest

import cloudinit.distros.bsd_utils as bsd_utils
from tests.unittests.helpers import mock

RC_FILE = """
if something; then
    do something here
fi
hostname={hostname}
"""


@mock.patch("cloudinit.distros.bsd_utils.util.load_text_file")
class TestBsdUtils:
    @pytest.mark.parametrize(
        "content,expected",
        (
            ("hostname=foo\n", "foo"),
            ("hostname=foo", "foo"),
            ('hostname="foo"', "foo"),
            ("hostname='foo'", "foo"),
            ("hostname='foo\"", "'foo\""),
            ("", None),
            (RC_FILE.format(hostname="foo"), "foo"),
        ),
    )
    def test_get_rc_config_value(self, m_load_file, content, expected):
        m_load_file.return_value = content
        assert bsd_utils.get_rc_config_value("hostname") == expected
        m_load_file.assert_called_with("/etc/rc.conf")

    @mock.patch("cloudinit.distros.bsd_utils.util.write_file")
    def test_set_rc_config_value_unchanged(self, m_write_file, m_load_file):
        m_load_file.return_value = RC_FILE.format(hostname="foo")
        m_write_file.assert_not_called()

    @mock.patch("cloudinit.distros.bsd_utils.util.write_file")
    def test_set_rc_config_value(self, m_write_file, m_load_file):
        bsd_utils.set_rc_config_value("hostname", "foo")
        m_write_file.assert_called_with("/etc/rc.conf", "hostname=foo\n")

        m_load_file.return_value = RC_FILE.format(hostname="foo")
        bsd_utils.set_rc_config_value("hostname", "bar")
        m_write_file.assert_called_with(
            "/etc/rc.conf", RC_FILE.format(hostname="bar")
        )



# This file is part of cloud-init. See LICENSE file for license information.

import logging

from cloudinit.distros import fetch
from cloudinit.subp import ProcessExecutionError
from tests.unittests.helpers import mock

M_PATH = "cloudinit.distros.raspberry_pi_os."


class TestRaspberryPiOS:
    @mock.patch("cloudinit.distros.debian.Distro.set_keymap")
    @mock.patch(M_PATH + "subp.subp")
    @mock.patch(M_PATH + "subp.which", return_value=False)
    def test_set_keymap_writes_file_and_runs_basics(
        self, m_which, m_subp, m_set_keymap
    ):
        cls = fetch("raspberry_pi_os")
        distro = cls("raspberry-pi-os", {}, None)

        args = [
            "gb",
            "pc105",
            "",
            "grp:alt_shift_toggle",
        ]

        # Avoid touching real services
        with mock.patch.object(cls, "manage_service") as m_manage_service:
            distro.set_keymap(*args)

            m_manage_service.assert_called_once_with(
                "restart", "keyboard-setup"
            )

        m_set_keymap.assert_called_once_with(*args)

        # Two raspi-config calls are always expected
        m_subp.assert_any_call(
            ["/usr/bin/raspi-config", "nonint", "update_labwc_keyboard"],
        )
        m_subp.assert_any_call(
            [
                "/usr/bin/raspi-config",
                "nonint",
                "update_squeekboard",
                "restart",
            ],
        )

        # No optional tools available, so only the two calls above
        assert m_subp.call_count == 2

    @mock.patch("cloudinit.distros.debian.Distro.set_keymap")
    @mock.patch(M_PATH + "subp.subp")
    def test_set_keymap_triggers_udevadm_when_available(
        self, m_subp, m_set_keymap
    ):
        cls = fetch("raspberry_pi_os")
        distro = cls("raspberry-pi-os", {}, None)

        # Only udevadm available
        def which_side_effect(name):
            return name == "udevadm"

        with mock.patch(M_PATH + "subp.which", side_effect=which_side_effect):
            with mock.patch.object(cls, "manage_service"):
                distro.set_keymap("de", "pc105", "nodeadkeys", "")

        m_set_keymap.assert_called_once_with("de", "pc105", "nodeadkeys", "")

        # Expect 3 subp calls: two raspi-config + one udevadm trigger
        assert m_subp.call_count == 3
        m_subp.assert_any_call(
            [
                "udevadm",
                "trigger",
                "--subsystem-match=input",
                "--action=change",
            ],
        )

    @mock.patch(M_PATH + "subp.subp")
    def test_apply_locale_happy_path(self, m_subp):
        cls = fetch("raspberry_pi_os")
        distro = cls("raspberry-pi-os", {}, None)
        distro.apply_locale("en_GB.UTF-8")
        m_subp.assert_called_once_with(
            [
                "/usr/bin/raspi-config",
                "nonint",
                "do_change_locale",
                "en_GB.UTF-8",
            ]
        )

    @mock.patch(M_PATH + "subp.subp")
    def test_apply_locale_fallback_to_utf8(self, m_subp):
        m_subp.side_effect = [
            ProcessExecutionError("Invalid locale"),  # Simulate failure
            None,  # Fallback succeeds
        ]
        cls = fetch("raspberry_pi_os")
        distro = cls("raspberry-pi-os", {}, None)
        distro.apply_locale("en_GB")
        assert m_subp.call_count == 2
        m_subp.assert_any_call(
            ["/usr/bin/raspi-config", "nonint", "do_change_locale", "en_GB"]
        )
        m_subp.assert_any_call(
            [
                "/usr/bin/raspi-config",
                "nonint",
                "do_change_locale",
                "en_GB.UTF-8",
            ]
        )

    @mock.patch(M_PATH + "subp.subp")
    def test_add_user_happy_path(self, m_subp):
        cls = fetch("raspberry_pi_os")
        distro = cls("raspberry-pi-os", {}, None)
        # Mock the superclass add_user to return True
        with mock.patch(
            "cloudinit.distros.debian.Distro.add_user", return_value=True
        ):
            assert distro.add_user("pi") is True
            m_subp.assert_called_once_with(
                ["/usr/bin/rename-user", "-f", "-s"],
                update_env={"SUDO_USER": "pi"},
            )

    @mock.patch(M_PATH + "subp.subp")
    def test_add_user_existing_user(self, m_subp):
        cls = fetch("raspberry_pi_os")
        distro = cls("raspberry-pi-os", {}, None)
        with mock.patch(
            "cloudinit.distros.debian.Distro.add_user", return_value=False
        ):
            assert distro.add_user("pi") is False
            m_subp.assert_not_called()

    @mock.patch(
        M_PATH + "subp.subp",
        side_effect=ProcessExecutionError("rename-user failed"),
    )
    @mock.patch("cloudinit.distros.debian.Distro.add_user", return_value=True)
    def test_add_user_rename_fails_logs_error(
        self, m_super_add_user, m_subp, caplog
    ):
        cls = fetch("raspberry_pi_os")
        distro = cls("raspberry-pi-os", {}, None)

        with caplog.at_level(logging.ERROR):
            assert distro.add_user("pi") is False
            assert "Failed to setup user" in caplog.text

    @mock.patch(
        "cloudinit.net.generate_fallback_config",
        return_value={"version": 1, "config": "fake"},
    )
    def test_fallback_netcfg(self, m_fallback_cfg, caplog):
        """Avoid fallback network unless disable_fallback_netcfg is False."""
        cls = fetch("raspberry_pi_os")
        distro = cls("raspberry-pi-os", {}, None)
        key = "disable_fallback_netcfg"
        expected_log_line = (
            "Skipping generation of fallback network "
            "config as per configuration. "
            "Rely on Raspberry Pi OS's default network configuration."
        )

        # The default skips fallback network config when no setting is given.
        caplog.clear()
        assert distro.generate_fallback_config() is None
        assert expected_log_line in caplog.text

        caplog.clear()
        distro._cfg[key] = True
        assert distro.generate_fallback_config() is None
        assert expected_log_line in caplog.text

        caplog.clear()
        distro._cfg[key] = False
        assert distro.generate_fallback_config() == {
            "version": 1,
            "config": "fake",
        }
        assert expected_log_line not in caplog.text



import os
import re
from contextlib import nullcontext
from copy import deepcopy
from textwrap import dedent
from typing import Any, Dict
from unittest import mock
from unittest.mock import MagicMock

from pytest import mark, param, raises

from cloudinit import lifecycle
from cloudinit.config import cc_ansible
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from cloudinit.subp import ProcessExecutionError
from tests.unittests.helpers import (
    skipUnlessJsonSchema,
    skipUnlessJsonSchemaVersionGreaterThan,
)
from tests.unittests.util import get_cloud

try:
    import pip as _pip  # noqa: F401

    HAS_PIP = True
except ImportError:
    HAS_PIP = False

M_PATH = "cloudinit.config.cc_ansible."
distro_version = dedent(
    """ansible 2.10.8
  config file = None
  configured module search path = ['/home/holmanb/.ansible/plugins/modules', \
  '/usr/share/ansible/plugins/modules']
  ansible python module location = /usr/lib/python3/dist-packages/ansible
  executable location = /usr/bin/ansible
  python version = 3.10.4 (main, Jun 29 2022, 12:14:53) [GCC 11.2.0]"""
)
pip_version = dedent(
    """ansible-pull [core 2.13.2]
  config file = None
  configured module search path = ['/root/.ansible/plugins/modules', \
  '/usr/share/ansible/plugins/modules']
  ansible python module location = /root/.local/lib/python3.8/site-\
  packages/ansible
  ansible collection location = /root/.ansible/collections:\
  /usr/share/ansible/collections
  executable location = /root/.local/lib/python3.8/site-packages/\
  ansible/__main__.py
  python version = 3.8.10 (default, Jun 22 2022, 20:18:18) [GCC 9.4.0]
  jinja version = 3.1.2
  libyaml = True """
)

CFG_CTRL = {
    "ansible": {
        "install_method": "distro",
        "package_name": "ansible-core",
        "ansible_config": "/etc/ansible/ansible.cfg",
        "galaxy": {
            "actions": [["ansible-galaxy", "install", "debops.apt"]],
        },
        "setup_controller": {
            "repositories": [
                {
                    "path": "/home/ansible/public/",
                    "source": "git@github.com:holmanb/ansible-lxd-public.git",
                },
                {
                    "path": "/home/ansible/private/",
                    "source": "git@github.com:holmanb/ansible-lxd-private.git",
                },
                {
                    "path": "/home/ansible/vmboot",
                    "source": "git@github.com:holmanb/vmboot.git",
                },
            ],
            "run_ansible": [
                {
                    "playbook_dir": "/home/ansible/my-repo",
                    "playbook_name": "start-lxd.yml",
                    "timeout": 120,
                    "forks": 1,
                    "private_key": "/home/ansible/.ssh/id_rsa",
                },
                {
                    "playbook_name": "configure-lxd.yml",
                    "become_user": "ansible",
                    "timeout": 120,
                    "forks": 1,
                    "private_key": "/home/ansible/.ssh/id_rsa",
                    "become_password_file": "/path/less/traveled",
                    "connection-password-file": "/path/more/traveled",
                    "module_path": "/path/head/traveled",
                    "vault_password_file": "/path/tail/traveled",
                    "playbook_dir": "/path/to/nowhere",
                    "inventory": "/a/file/as/well",
                },
            ],
        },
    },
}

CFG_FULL_PULL_DICT: Dict[str, Any] = {
    "ansible": {
        "install_method": "distro",
        "package_name": "ansible-core",
        "ansible_config": "/etc/ansible/ansible.cfg",
        "galaxy": {
            "actions": [["ansible-galaxy", "install", "debops.apt"]],
        },
        "pull": {
            "url": "https://github/holmanb/vmboot",
            "playbook_names": ["arch.yml"],
            "accept_host_key": True,
            "clean": True,
            "full": True,
            "diff": False,
            "ssh_common_args": "-y",
            "scp_extra_args": "-l",
            "sftp_extra_args": "-f",
            "checkout": "tree",
            "module_path": "~/.ansible/plugins/modules:"
            "/usr/share/ansible/plugins/modules",
            "timeout": "10",
            "vault_id": "me",
            "connection": "smart",
            "vault_password_file": "/path/to/file",
            "module_name": "git",
            "sleep": "1",
            "tags": "cumulus",
            "skip_tags": "cisco",
            "private_key": "{nope}",
        },
    }
}

CFG_FULL_PULL_LIST: Dict[str, Any] = {
    "ansible": {
        "install_method": "distro",
        "package_name": "ansible-core",
        "ansible_config": "/etc/ansible/ansible.cfg",
        "galaxy": {
            "actions": [["ansible-galaxy", "install", "debops.apt"]],
        },
        "pull": [
            {
                "url": "https://github/holmanb/vmboot",
                "playbook_names": ["arch.yml"],
                "accept_host_key": True,
                "clean": True,
                "full": True,
                "diff": False,
                "ssh_common_args": "-y",
                "scp_extra_args": "-l",
                "sftp_extra_args": "-f",
                "checkout": "tree",
                "module_path": "~/.ansible/plugins/modules:"
                "/usr/share/ansible/plugins/modules",
                "timeout": "10",
                "vault_id": "me",
                "connection": "smart",
                "vault_password_file": "/path/to/file",
                "module_name": "git",
                "sleep": "1",
                "tags": "cumulus",
                "skip_tags": "cisco",
                "private_key": "{nope}",
            }
        ],
    }
}

CFG_MINIMAL_DICT = {
    "ansible": {
        "install_method": "pip",
        "package_name": "ansible",
        "run_user": "ansible",
        "pull": {
            "url": "https://github/holmanb/vmboot",
            "playbook_names": ["ubuntu.yml"],
        },
    }
}

CFG_MINIMAL_LIST = {
    "ansible": {
        "install_method": "pip",
        "package_name": "ansible",
        "run_user": "ansible",
        "pull": [
            {
                "url": "https://github/holmanb/vmboot",
                "playbook_names": ["ubuntu.yml"],
            }
        ],
    }
}


class TestSchema:

    @mark.parametrize(
        ("config", "error_msg"),
        (
            param(
                CFG_MINIMAL_DICT,
                "Expect **ansible.pull** as list of objects",
                id="essentials_dict",
            ),
            param(
                CFG_FULL_PULL_DICT,
                "Expect **ansible.pull** as list of objects",
                id="all-pull-keys-dict",
            ),
        ),
    )
    @skipUnlessJsonSchemaVersionGreaterThan(version=(3, 2, 0))
    def test_schema_validation_deprecations(self, config, error_msg):
        with raises(SchemaValidationError, match=re.escape(error_msg)):
            validate_cloudconfig_schema(config, get_schema(), strict=True)

    @mark.parametrize(
        ("config", "error_msg"),
        (
            param(
                {
                    "ansible": {
                        "install_method": "distro",
                        "pull": {
                            "url": "https://github/holmanb/vmboot",
                            "playbook_names": ["centos.yml"],
                            "dance": "bossa nova",
                        },
                    }
                },
                "Additional properties are not allowed ",
                id="additional-properties-dict",
            ),
            param(
                CFG_CTRL,
                None,
                id="ctrl-keys",
            ),
            param(
                {
                    "ansible": {
                        "install_method": "true",
                        "pull": {
                            "url": "https://github/holmanb/vmboot",
                            "playbook_names": ["debian.yml"],
                        },
                    }
                },
                "'true' is not one of ['distro', 'pip']",
                id="install-type-dict",
            ),
            param(
                {
                    "ansible": {
                        "install_method": "pip",
                        "pull": {
                            "playbook_names": ["fedora.yml"],
                        },
                    }
                },
                "'url' is a required property",
                id="require-url-dict",
            ),
            param(
                {
                    "ansible": {
                        "install_method": "pip",
                        "pull": {
                            "url": "gophers://encrypted-gophers/",
                        },
                    }
                },
                "'playbook_name' is a required property",
                id="require-url-dict",
            ),
            param(
                CFG_MINIMAL_LIST,
                None,
                id="essentials_list",
            ),
            param(
                {
                    "ansible": {
                        "install_method": "distro",
                        "pull": [
                            {
                                "url": "https://github/holmanb/vmboot",
                                "playbook_names": ["centos.yml"],
                                "dance": "bossa nova",
                            }
                        ],
                    }
                },
                "Additional properties are not allowed ",
                id="additional-properties-list",
            ),
            param(
                CFG_FULL_PULL_LIST,
                None,
                id="all-pull-keys-list",
            ),
            param(
                CFG_CTRL,
                None,
                id="ctrl-keys",
            ),
            param(
                {
                    "ansible": {
                        "install_method": "true",
                        "pull": [
                            {
                                "url": "https://github/holmanb/vmboot",
                                "playbook_names": ["debian.yml"],
                            }
                        ],
                    }
                },
                "'true' is not one of ['distro', 'pip']",
                id="install-type-list",
            ),
            param(
                {
                    "ansible": {
                        "install_method": "pip",
                        "pull": [
                            {
                                "playbook_names": ["fedora.yml"],
                            }
                        ],
                    }
                },
                "'url' is a required property",
                id="require-url-list",
            ),
            param(
                {
                    "ansible": {
                        "install_method": "pip",
                        "pull": [
                            {
                                "url": "gophers://encrypted-gophers/",
                            }
                        ],
                    }
                },
                "'playbook_name' is a required property",
                id="require-url-list",
            ),
        ),
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with raises(SchemaValidationError, match=re.escape(error_msg)):
                validate_cloudconfig_schema(config, get_schema(), strict=True)


class TestAnsible:
    def test_filter_args(self):
        """only diff should be removed"""
        src_cfg = deepcopy(
            CFG_FULL_PULL_DICT.get("ansible", {}).get("pull", {})
        )
        src_cfg.pop("playbook_names")
        out_dict = cc_ansible.filter_args(src_cfg)
        assert out_dict == {
            "url": "https://github/holmanb/vmboot",
            "accept-host-key": True,
            "clean": True,
            "full": True,
            "ssh-common-args": "-y",
            "scp-extra-args": "-l",
            "sftp-extra-args": "-f",
            "checkout": "tree",
            "module-path": "~/.ansible/plugins/modules:"
            "/usr/share/ansible/plugins/modules",
            "timeout": "10",
            "vault-id": "me",
            "connection": "smart",
            "vault-password-file": "/path/to/file",
            "module-name": "git",
            "sleep": "1",
            "tags": "cumulus",
            "skip-tags": "cisco",
            "private-key": "{nope}",
        }

    @mark.parametrize(
        ("cfg_dict", "exception"),
        (
            (CFG_FULL_PULL_DICT, None),
            (CFG_MINIMAL_DICT, None),
            (
                {
                    "ansible": {
                        "package_name": "ansible-core",
                        "install_method": "distro",
                        "pull": {
                            "playbook_names": ["ubuntu.yml"],
                        },
                    }
                },
                ValueError,
            ),
            (
                {
                    "ansible": {
                        "install_method": "pip",
                        "pull": {
                            "url": "https://github/holmanb/vmboot",
                        },
                    }
                },
                ValueError,
            ),
        ),
    )
    def test_required_keys_dict(self, cfg_dict, exception, mocker):
        mocker.patch(M_PATH + "subp.subp", return_value=("", ""))
        mocker.patch(M_PATH + "subp.which", return_value=True)
        mocker.patch(M_PATH + "AnsiblePull.check_deps")
        mocker.patch(
            M_PATH + "AnsiblePull.get_version",
            return_value=cc_ansible.lifecycle.Version(2, 7, 1),
        )
        mocker.patch(
            M_PATH + "AnsiblePullDistro.is_installed",
            return_value=False,
        )
        if exception:
            with raises(exception):
                cc_ansible.handle("", cfg_dict, get_cloud(), [])
        else:
            cloud = get_cloud(mocked_distro=True)
            cloud.distro.pip_package_name = "python3-pip"
            install = cfg_dict["ansible"]["install_method"]
            cc_ansible.handle("", cfg_dict, cloud, [])
            if install == "distro":
                cloud.distro.install_packages.assert_called_once()
                cloud.distro.install_packages.assert_called_with(
                    ["ansible-core"]
                )
            elif install == "pip":
                if HAS_PIP:
                    assert 0 == cloud.distro.install_packages.call_count
                else:
                    cloud.distro.install_packages.assert_called_with(
                        ["python3-pip"]
                    )

    @mark.parametrize(
        ("cfg_list", "exception"),
        (
            (CFG_FULL_PULL_LIST, None),
            (CFG_MINIMAL_LIST, None),
            (
                {
                    "ansible": {
                        "package_name": "ansible-core",
                        "install_method": "distro",
                        "pull": [
                            {
                                "playbook_names": ["ubuntu.yml"],
                            }
                        ],
                    }
                },
                ValueError,
            ),
            (
                {
                    "ansible": {
                        "install_method": "pip",
                        "pull": [
                            {
                                "url": "https://github/holmanb/vmboot",
                            }
                        ],
                    }
                },
                ValueError,
            ),
        ),
    )
    def test_required_keys_list(self, cfg_list, exception, mocker):
        mocker.patch(M_PATH + "subp.subp", return_value=("", ""))
        mocker.patch(M_PATH + "subp.which", return_value=True)
        mocker.patch(M_PATH + "AnsiblePull.check_deps")
        mocker.patch(
            M_PATH + "AnsiblePull.get_version",
            return_value=cc_ansible.lifecycle.Version(2, 7, 1),
        )
        mocker.patch(
            M_PATH + "AnsiblePullDistro.is_installed",
            return_value=False,
        )
        if exception:
            with raises(exception):
                cc_ansible.handle("", cfg_list, get_cloud(), [])
        else:
            cloud = get_cloud(mocked_distro=True)
            cloud.distro.pip_package_name = "python3-pip"
            install = cfg_list["ansible"]["install_method"]
            cc_ansible.handle("", cfg_list, cloud, [])
            if install == "distro":
                cloud.distro.install_packages.assert_called_once()
                cloud.distro.install_packages.assert_called_with(
                    ["ansible-core"]
                )
            elif install == "pip":
                if HAS_PIP:
                    assert 0 == cloud.distro.install_packages.call_count
                else:
                    cloud.distro.install_packages.assert_called_with(
                        ["python3-pip"]
                    )

    @mock.patch(M_PATH + "subp.which", return_value=False)
    def test_deps_not_installed(self, m_which):
        """assert exception raised if package not installed"""
        with raises(ValueError):
            cc_ansible.AnsiblePullDistro(
                get_cloud().distro, "root"
            ).check_deps()

    @mock.patch(M_PATH + "subp.which", return_value=True)
    def test_deps(self, m_which):
        """assert exception not raised if package installed"""
        cc_ansible.AnsiblePullDistro(
            get_cloud().distro, "ansible"
        ).check_deps()

    @mark.serial
    @mock.patch(M_PATH + "subp.subp", return_value=("stdout", "stderr"))
    @mock.patch(M_PATH + "subp.which", return_value=False)
    def test_pip_bootstrap(self, m_which, m_subp):
        distro = get_cloud(mocked_distro=True).distro
        with mock.patch("builtins.__import__", side_effect=ImportError):
            cc_ansible.AnsiblePullPip(distro, "ansible").install("")
        distro.install_packages.assert_called_once()

    @mark.parametrize(
        "mocked_doas_kwargs,expected_result,expected_present_logs,expected_absent_logs",
        [
            (
                {"return_value": ("", "")},
                nullcontext(),
                ["Upgraded pip"],
                ["Failed at upgrading pip"],
            ),
            # When do_as raises a ProcessExecutionError, __upgrade_pip will
            #  catch it then return without raising it
            (
                {"side_effect": ProcessExecutionError()},
                nullcontext(),
                ["Failed at upgrading pip"],
                ["Upgraded pip"],
            ),
            # The below use case is to ensure that the function will raise the
            # error if the error is not ProcessExecutionError (i.e. something
            # went really sideways with do_as and not simply a non-zero exit
            # by the executed command)
            (
                {"side_effect": ZeroDivisionError()},
                raises(ZeroDivisionError),
                [],
                ["Failed at upgrading pip", "Upgraded pip"],
            ),
        ],
    )
    def test_upgrade_pip(
        self,
        mocked_doas_kwargs,
        expected_result,
        expected_present_logs,
        expected_absent_logs,
        caplog,
    ):
        """assert __upgrade_pip will succeed
        or will catch an exception and not fail"""
        distro = get_cloud(mocked_distro=True).distro
        pip = cc_ansible.AnsiblePullPip(distro, "ansible")
        base_cmd = ["echo", "abc"]
        with mock.patch.object(
            pip, "do_as", **mocked_doas_kwargs
        ) as mocked_doas:
            with expected_result as e:
                assert (
                    getattr(pip, "_AnsiblePullPip__upgrade_pip")(base_cmd) == e
                )

                mocked_doas.assert_called_once()
                mocked_doas.assert_called_with([*base_cmd, "--upgrade", "pip"])

                for expected_present_log in expected_present_logs:
                    assert expected_present_log in caplog.text

                for expected_absent_log in expected_absent_logs:
                    assert expected_absent_log not in caplog.text

    @mock.patch(M_PATH + "subp.which", return_value=True)
    @mock.patch(M_PATH + "subp.subp", return_value=(distro_version, "stderr"))
    @mock.patch(
        "cloudinit.distros.subp.subp", return_value=(distro_version, "stderr")
    )
    @mark.parametrize(
        ("cfg_dict", "expected"),
        (
            (
                CFG_FULL_PULL_DICT,
                [
                    "ansible-pull",
                    "--url=https://github/holmanb/vmboot",
                    "--accept-host-key",
                    "--clean",
                    "--full",
                    "--ssh-common-args=-y",
                    "--scp-extra-args=-l",
                    "--sftp-extra-args=-f",
                    "--checkout=tree",
                    "--module-path=~/.ansible/plugins/modules"
                    ":/usr/share/ansible/plugins/modules",
                    "--timeout=10",
                    "--vault-id=me",
                    "--connection=smart",
                    "--vault-password-file=/path/to/file",
                    "--module-name=git",
                    "--sleep=1",
                    "--tags=cumulus",
                    "--skip-tags=cisco",
                    "--private-key={nope}",
                    "arch.yml",
                ],
            ),
            (
                CFG_MINIMAL_DICT,
                [
                    "ansible-pull",
                    "--url=https://github/holmanb/vmboot",
                    "ubuntu.yml",
                ],
            ),
        ),
    )
    def test_ansible_pull_dict(
        self, m_subp1, m_subp2, m_which, cfg_dict, expected
    ):
        """verify expected ansible invocation from userdata config"""
        pull_type = cfg_dict["ansible"]["install_method"]
        distro = get_cloud().distro
        distro.do_as = MagicMock(return_value=(distro_version, ""))

        ansible_pull = (
            cc_ansible.AnsiblePullPip(distro, "ansible")
            if pull_type == "pip"
            else cc_ansible.AnsiblePullDistro(distro, "")
        )
        cc_ansible.run_ansible_pull(
            ansible_pull, deepcopy(cfg_dict["ansible"]["pull"])
        )

        if pull_type != "pip":
            assert m_subp2.call_args[0][0] == expected
            assert m_subp2.call_args[1]["update_env"].get(
                "HOME"
            ) == os.environ.get("HOME", "/root")

    @mock.patch(M_PATH + "subp.which", return_value=True)
    @mock.patch(M_PATH + "subp.subp", return_value=(distro_version, ""))
    @mock.patch(
        "cloudinit.distros.subp.subp", return_value=(distro_version, "")
    )
    @mark.parametrize(
        ("cfg_list", "expected"),
        (
            (
                CFG_FULL_PULL_LIST,
                [
                    "ansible-pull",
                    "--url=https://github/holmanb/vmboot",
                    "--accept-host-key",
                    "--clean",
                    "--full",
                    "--ssh-common-args=-y",
                    "--scp-extra-args=-l",
                    "--sftp-extra-args=-f",
                    "--checkout=tree",
                    "--module-path=~/.ansible/plugins/modules"
                    ":/usr/share/ansible/plugins/modules",
                    "--timeout=10",
                    "--vault-id=me",
                    "--connection=smart",
                    "--vault-password-file=/path/to/file",
                    "--module-name=git",
                    "--sleep=1",
                    "--tags=cumulus",
                    "--skip-tags=cisco",
                    "--private-key={nope}",
                    "arch.yml",
                ],
            ),
            (
                CFG_MINIMAL_LIST,
                [
                    "ansible-pull",
                    "--url=https://github/holmanb/vmboot",
                    "ubuntu.yml",
                ],
            ),
        ),
    )
    def test_ansible_pull_list(
        self, m_subp1, m_subp2, m_which, cfg_list, expected
    ):
        """verify expected ansible invocation from userdata config"""
        pull_type = cfg_list["ansible"]["install_method"]
        distro = get_cloud().distro
        distro.do_as = MagicMock(return_value=(distro_version, ""))
        ansible_pull = (
            cc_ansible.AnsiblePullPip(distro, "ansible")
            if pull_type == "pip"
            else cc_ansible.AnsiblePullDistro(distro, "")
        )
        for cfg in cfg_list["ansible"]["pull"]:
            cc_ansible.run_ansible_pull(ansible_pull, deepcopy(cfg))

        if pull_type != "pip":
            assert m_subp2.call_args[0][0] == expected
            assert m_subp2.call_args[1]["update_env"].get(
                "HOME"
            ) == os.environ.get("HOME", "/root")

    @mock.patch(M_PATH + "validate_config")
    def test_do_not_run(self, m_validate):
        """verify that if ansible key not included, don't do anything"""
        cc_ansible.handle("", {}, get_cloud(), [])
        assert not m_validate.called

    @mock.patch(
        "cloudinit.config.cc_ansible.subp.subp",
        side_effect=[(distro_version, "")],
    )
    def test_parse_version_distro(self, m_subp):
        """Verify that the expected version is returned"""
        assert cc_ansible.AnsiblePullDistro(
            get_cloud().distro, ""
        ).get_version() == lifecycle.Version(2, 10, 8)

    @mock.patch("cloudinit.subp.subp", side_effect=[(pip_version, "")])
    def test_parse_version_pip(self, m_subp):
        """Verify that the expected version is returned"""
        distro = get_cloud().distro
        distro.do_as = MagicMock(return_value=(pip_version, ""))
        pip = cc_ansible.AnsiblePullPip(distro, "root")
        received = pip.get_version()
        expected = lifecycle.Version(2, 13, 2)
        assert received == expected

    @mock.patch(
        M_PATH + "subp.subp",
        side_effect=[("out", "err"), (distro_version, ""), ("out", "err")],
    )
    @mock.patch(M_PATH + "subp.which", return_value=True)
    def test_ansible_env_var(self, m_which, m_subp):
        cc_ansible.handle("", CFG_FULL_PULL_DICT, get_cloud(), [])

        # python 3.8 required for Mock.call_args.kwargs dict attribute
        if isinstance(m_subp.call_args.kwargs, dict):
            assert (
                "/etc/ansible/ansible.cfg"
                == m_subp.call_args.kwargs["update_env"]["ansible_config"]
            )



# This file is part of cloud-init. See LICENSE file for license information.

from tests.unittests.helpers import get_distro

SYSTEM_INFO = {
    "paths": {
        "cloud_dir": "/var/lib/cloud/",
        "templates_dir": "/etc/cloud/templates/",
    },
    "network": {"renderers": "networkd"},
}


class TestAzurelinux:
    distro = get_distro("azurelinux", SYSTEM_INFO)
    expected_log_line = "Rely on Azure Linux default network config"

    def test_network_renderer(self):
        assert self.distro._cfg["network"]["renderers"] == "networkd"

    def test_get_distro(self):
        assert self.distro.osfamily == "azurelinux"



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
import inspect
import logging
import signal
import sys
import types
from io import StringIO
from typing import Callable, Dict, Final, Union

from cloudinit import version as vr
from cloudinit.log import log_util

LOG = logging.getLogger(__name__)

SIG_MESSAGE: Final = "Cloud-init {} received {}, exiting\n"
BACK_FRAME_TRACE_DEPTH: Final = 3
SIGNALS: Final[Dict[int, str]] = {
    signal.SIGINT: "Cloud-init %(version)s received SIGINT, exiting",
    signal.SIGTERM: "Cloud-init %(version)s received SIGTERM, exiting",
    signal.SIGABRT: "Cloud-init %(version)s received SIGABRT, exiting",
}


def inspect_handler(sig: Union[int, Callable, None]) -> None:
    """inspect_handler() logs signal handler state"""
    if callable(sig):
        # only produce a log when the signal handler isn't in the expected
        # default state
        if not isinstance(sig, types.BuiltinFunctionType):
            LOG.info("Signal state [%s] - previously custom handler.", sig)
    elif sig == signal.SIG_IGN:
        LOG.info("Signal state [SIG_IGN] - previously ignored.")
    elif sig is None:
        LOG.info("Signal state [None] - previously not installed from Python.")
    elif sig == signal.SIG_DFL:
        LOG.info(
            "Signal state [%s] - default way of handling signal was "
            "previously in use.",
            sig,
        )
    else:
        # this should never happen, unless something in Python changes
        # https://docs.python.org/3/library/signal.html#signal.getsignal
        LOG.warning("Signal state [%s(%s)] - unknown", type(sig), sig)


def _pprint_frame(frame, depth, max_depth, contents):
    if depth > max_depth or not frame:
        return
    frame_info = inspect.getframeinfo(frame)
    prefix = " " * (depth * 2)
    contents.write("%sFilename: %s\n" % (prefix, frame_info.filename))
    contents.write("%sFunction: %s\n" % (prefix, frame_info.function))
    contents.write("%sLine number: %s\n" % (prefix, frame_info.lineno))
    _pprint_frame(frame.f_back, depth + 1, max_depth, contents)


def _handle_exit(signum, frame):
    # in practice we always receive a Signals object but int is possible
    name = signum.name if isinstance(signum, signal.Signals) else signum
    contents = StringIO(SIG_MESSAGE.format(vr.version_string(), name))
    _pprint_frame(frame, 1, BACK_FRAME_TRACE_DEPTH, contents)
    log_util.multi_log(
        f"Received signal {name} resulting in exit. Cause:\n"
        + contents.getvalue(),
        log=LOG,
        log_level=logging.INFO,
    )
    sys.exit(0)


def attach_handlers():
    """attach cloud-init's handlers"""
    sigs_attached = 0
    for signum in SIGNALS.keys():
        inspect_handler(signal.signal(signum, _handle_exit))
    sigs_attached += len(SIGNALS)
    return sigs_attached



import functools
import logging
import time

LOG = logging.getLogger(__name__)


class Timed:
    """
    A context manager which measures and optionally logs context run time.

    :param msg: A message that describes the thing that is being measured
    :param threshold: Threshold, in seconds. When the context exceeds this
        threshold, a log will be made.
    :param log_mode: Control whether to log. Defaults to "threshold". Possible
        values include:
        "always" - Always log 'msg', even when 'threshold' is not reached.
        "threshold" - Log when context time exceeds 'threshold'.
        "skip" - Do not log. Context time and message are stored in the
            'output' and 'delta' attributes, respectively. Used to manually
            coalesce with other logs at the call site.

    usage:

        this call:
        ```
        with Timed("Configuring the network"):
            run_configure()
        ```

        might produce this log:
        ```
            Configuring the network took 0.100 seconds
        ```
    """

    def __init__(
        self,
        msg: str,
        *,
        threshold: float = 0.01,
        log_mode: str = "threshold",
    ):
        self.msg = msg
        self.threshold = threshold
        self.log_mode = log_mode
        self.output = ""
        self.start = 0.0
        self.delta = 0.0

    def __enter__(self):
        self.start = time.monotonic()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delta = time.monotonic() - self.start
        suffix = f"took {self.delta:.3f} seconds"
        if "always" == self.log_mode:
            LOG.debug("%s %s", self.msg, suffix)
        elif "skip" == self.log_mode:
            return
        elif "threshold" == self.log_mode:
            if self.delta > self.threshold:
                LOG.debug("%s %s", self.msg, suffix)
                self.output = f"{self.msg} {suffix}"
        else:
            raise ValueError(
                f"Invalid Timed log_mode value: '{self.log_mode}'."
            )


def timed(msg: str, *, threshold: float = 0.01, log_mode: str = "threshold"):
    """
    A decorator which measures and optionally logs context run time.

    :param msg: A message that describes the thing that is being measured
    :param threshold: Threshold, in seconds. When the context exceeds this
        threshold, a log will be made.
    :param log_mode: Control whether to log. Defaults to "threshold". Possible
        values include:
        "always" - Always log 'msg', even when 'threshold' is not reached.
        "threshold" - Log when context time exceeds 'threshold'.

    usage:

        this call:
        ```
        @timed("Configuring the network")
        def run_configure():
            ...
        ```

        might produce this log:
        ```
            Configuring the network took 0.100 seconds
        ```
    """

    def wrapper(func):
        @functools.wraps(func)
        def decorator(*args, **kwargs):
            with Timed(msg, threshold=threshold, log_mode=log_mode):
                return func(*args, **kwargs)

        return decorator

    return wrapper



# Copyright (C) 2015 Canonical Ltd.
# Copyright (C) 2006-2024 Broadcom. All Rights Reserved.
# Broadcom Confidential. The term "Broadcom" refers to Broadcom Inc.
# and/or its subsidiaries.
#
# Author: Sankar Tanguturi <stanguturi@vmware.com>
#         Pengpeng Sun <pengpeng.sun@broadcom.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import ipaddress
import logging
import os
import re

from cloudinit import net, subp, util
from cloudinit.net.network_state import (
    ipv4_mask_to_net_prefix,
    ipv6_mask_to_net_prefix,
)

logger = logging.getLogger(__name__)


class NicConfigurator:
    def __init__(
        self, nics, name_servers, dns_suffixes, use_system_devices=True
    ):
        """
        Initialize the Nic Configurator
        @param nics (list) an array of nics to configure
        @param use_system_devices (Bool) Get the MAC names from the system
        if this is True. If False, then mac names will be retrieved from
         the specified nics.
        """
        self.nics = nics
        self.name_servers = name_servers
        self.dns_suffixes = dns_suffixes
        self.mac2Name = {}

        if use_system_devices:
            self.find_devices()
        else:
            for nic in self.nics:
                self.mac2Name[nic.mac.lower()] = nic.name

        self._primaryNic = self.get_primary_nic()

    def get_primary_nic(self):
        """
        Retrieve the primary nic if it exists
        @return (NicBase): the primary nic if exists, None otherwise
        """
        primary_nics = [nic for nic in self.nics if nic.primary]
        if not primary_nics:
            return None
        elif len(primary_nics) > 1:
            raise RuntimeError(
                "There can only be one primary nic",
                [nic.mac for nic in primary_nics],
            )
        else:
            return primary_nics[0]

    def find_devices(self):
        """
        Create the mac2Name dictionary
        The mac address(es) are in the lower case
        """
        cmd = ["ip", "addr", "show"]
        output, _err = subp.subp(cmd)
        sections = re.split(r"\n\d+: ", "\n" + output)[1:]

        macPat = r"link/ether (([0-9A-Fa-f]{2}[:]){5}([0-9A-Fa-f]{2}))"
        for section in sections:
            match = re.search(macPat, section)
            if not match:  # Only keep info about nics
                continue
            mac = match.group(1).lower()
            name = section.split(":", 1)[0]
            self.mac2Name[mac] = name

    def gen_one_nic_v2(self, nic):
        """
        Return the config dict needed to configure a nic
        @return (dict): the config dict to configure the nic
        @param nic (NicBase): the nic to configure
        """
        mac = nic.mac.lower()
        name = self.mac2Name.get(mac)
        if not name:
            raise ValueError("No known device has MACADDR: %s" % nic.mac)

        nic_config_dict = {}
        generators = [
            self.gen_match(mac),
            self.gen_set_name(name),
            self.gen_wakeonlan(nic),
            self.gen_dhcp4(nic),
            self.gen_dhcp6(nic),
            self.gen_addresses(nic),
            self.gen_routes(nic),
            self.gen_nameservers(),
        ]
        for value in generators:
            if value:
                nic_config_dict.update(value)

        return {name: nic_config_dict}

    def gen_match(self, mac):
        return {"match": {"macaddress": mac}}

    def gen_set_name(self, name):
        return {"set-name": name}

    def gen_wakeonlan(self, nic):
        return {"wakeonlan": nic.onboot}

    def gen_dhcp4(self, nic):
        dhcp4 = {}
        bootproto = nic.bootProto.lower()
        if nic.ipv4_mode.lower() == "disabled":
            bootproto = "manual"
        if bootproto != "static":
            dhcp4.update({"dhcp4": True})
            # dhcp4-overrides
            if self.name_servers or self.dns_suffixes:
                dhcp4.update({"dhcp4-overrides": {"use-dns": False}})
        else:
            dhcp4.update({"dhcp4": False})
        return dhcp4

    def gen_dhcp6(self, nic):
        dhcp6 = {}
        if nic.staticIpv6:
            dhcp6.update({"dhcp6": False})
        # TODO: nic shall explicitly tell it's DHCP6
        # TODO: set dhcp6-overrides
        return dhcp6

    def gen_addresses(self, nic):
        address_list = []
        v4_cidr = 32

        # Static Ipv4
        v4_addrs = nic.staticIpv4
        if v4_addrs:
            v4 = v4_addrs[0]
            if v4.netmask:
                v4_cidr = ipv4_mask_to_net_prefix(v4.netmask)
            if v4.ip:
                address_list.append(f"{v4.ip}/{v4_cidr}")
        # Static Ipv6
        v6_addrs = nic.staticIpv6
        if v6_addrs:
            for v6 in v6_addrs:
                v6_cidr = ipv6_mask_to_net_prefix(v6.netmask)
                address_list.append(f"{v6.ip}/{v6_cidr}")

        if address_list:
            return {"addresses": address_list}
        else:
            return {}

    def gen_routes(self, nic):
        route_list = []
        v4_cidr = 32

        # Ipv4 routes
        v4_addrs = nic.staticIpv4
        if v4_addrs:
            v4 = v4_addrs[0]
            # Add the ipv4 default route
            if nic.primary and v4.gateways:
                route_list.append({"to": "0.0.0.0/0", "via": v4.gateways[0]})
            # Add ipv4 static routes if there is no primary nic
            if not self._primaryNic and v4.gateways:
                if v4.netmask:
                    v4_cidr = ipv4_mask_to_net_prefix(v4.netmask)
                for gateway in v4.gateways:
                    v4_subnet = ipaddress.IPv4Network(
                        f"{gateway}/{v4_cidr}", strict=False
                    )
                    route_list.append({"to": f"{v4_subnet}", "via": gateway})
        # Ipv6 routes
        v6_addrs = nic.staticIpv6
        if v6_addrs:
            for v6 in v6_addrs:
                v6_cidr = ipv6_mask_to_net_prefix(v6.netmask)
                # Add the ipv6 default route
                if nic.primary and v6.gateway:
                    route_list.append({"to": "::/0", "via": v6.gateway})
                # Add ipv6 static routes if there is no primary nic
                if not self._primaryNic and v6.gateway:
                    v6_subnet = ipaddress.IPv6Network(
                        f"{v6.gateway}/{v6_cidr}", strict=False
                    )
                    route_list.append(
                        {"to": f"{v6_subnet}", "via": v6.gateway}
                    )

        if route_list:
            return {"routes": route_list}
        else:
            return {}

    def gen_nameservers(self):
        nameservers_dict = {}
        search_list = []
        addresses_list = []
        if self.dns_suffixes:
            for dns_suffix in self.dns_suffixes:
                search_list.append(dns_suffix)
        if self.name_servers:
            for name_server in self.name_servers:
                addresses_list.append(name_server)
        if search_list:
            nameservers_dict.update({"search": search_list})
        if addresses_list:
            nameservers_dict.update({"addresses": addresses_list})

        if nameservers_dict:
            return {"nameservers": nameservers_dict}
        else:
            return {}

    def generate(self, configure=False, osfamily=None):
        """Return the config elements that are needed to configure the nics"""
        if configure:
            logger.info("Configuring the interfaces file")
            self.configure(osfamily)

        ethernets_dict = {}

        for nic in self.nics:
            ethernets_dict.update(self.gen_one_nic_v2(nic))

        return ethernets_dict

    def clear_dhcp(self):
        logger.info("Clearing DHCP leases")
        net.dhcp.IscDhclient.clear_leases()

    def configure(self, osfamily=None):
        """
        Configure the /etc/network/interfaces
        Make a back up of the original
        """

        if not osfamily or osfamily != "debian":
            logger.info("Debian OS not detected. Skipping the configure step")
            return

        containingDir = "/etc/network"

        interfaceFile = os.path.join(containingDir, "interfaces")
        originalFile = os.path.join(
            containingDir, "interfaces.before_vmware_customization"
        )

        if not os.path.exists(originalFile) and os.path.exists(interfaceFile):
            os.rename(interfaceFile, originalFile)

        lines = [
            "# DO NOT EDIT THIS FILE BY HAND --"
            " AUTOMATICALLY GENERATED BY cloud-init",
            "source /etc/network/interfaces.d/*",
            "source-directory /etc/network/interfaces.d",
        ]

        util.write_file(interfaceFile, content="\n".join(lines))

        self.clear_dhcp()



# This file is part of cloud-init. See LICENSE file for license information.

"""Tests for cloudinit.subp utility functions"""

import json
import logging
import os
import stat
import sys
from unittest import mock

import pytest

from cloudinit import subp, util
from tests.helpers import get_top_level_dir

SH = "sh"
BOGUS_COMMAND = "this-is-not-expected-to-be-a-program-name"


class TestPrependBaseCommands:

    def test_prepend_base_command_errors_on_neither_string_nor_list(self):
        """Raise an error for each command which is not a string or list."""
        orig_commands = ["ls", 1, {"not": "gonna work"}, ["basecmd", "list"]]
        with pytest.raises(
            TypeError,
            match="Invalid basecmd config. These commands are not a string or"
            " list:\n1\n{'not': 'gonna work'}",
        ):
            subp.prepend_base_command(
                base_command="basecmd", commands=orig_commands
            )

    def test_prepend_base_command_warns_on_non_base_string_commands(
        self, caplog
    ):
        """Warn on each non-base for commands of type string."""
        orig_commands = [
            "ls",
            "basecmd list",
            "touch /blah",
            "basecmd install x",
        ]
        fixed_commands = subp.prepend_base_command(
            base_command="basecmd", commands=orig_commands
        )
        assert (
            mock.ANY,
            logging.WARNING,
            "Non-basecmd commands in basecmd config:\nls\ntouch /blah",
        ) in caplog.record_tuples
        assert orig_commands == fixed_commands

    def test_prepend_base_command_prepends_on_non_base_list_commands(
        self, caplog
    ):
        """Prepend 'basecmd' for each non-basecmd command of type list."""
        orig_commands = [
            ["ls"],
            ["basecmd", "list"],
            ["basecmda", "/blah"],
            ["basecmd", "install", "x"],
        ]
        expected = [
            ["basecmd", "ls"],
            ["basecmd", "list"],
            ["basecmd", "basecmda", "/blah"],
            ["basecmd", "install", "x"],
        ]
        fixed_commands = subp.prepend_base_command(
            base_command="basecmd", commands=orig_commands
        )
        assert "" == caplog.text
        assert expected == fixed_commands

    def test_prepend_base_command_removes_first_item_when_none(self, caplog):
        """Remove the first element of a non-basecmd when it is None."""
        orig_commands = [
            [None, "ls"],
            ["basecmd", "list"],
            [None, "touch", "/blah"],
            ["basecmd", "install", "x"],
        ]
        expected = [
            ["ls"],
            ["basecmd", "list"],
            ["touch", "/blah"],
            ["basecmd", "install", "x"],
        ]
        fixed_commands = subp.prepend_base_command(
            base_command="basecmd", commands=orig_commands
        )
        assert "" == caplog.text
        assert expected == fixed_commands


@pytest.mark.allow_all_subp
class TestSubp:
    stdin2err = [SH, "-c", "cat >&2"]
    stdin2out = ["cat"]
    utf8_invalid = b"ab\xaadef"
    utf8_valid = b"start \xc3\xa9 end"
    utf8_valid_2 = b"d\xc3\xa9j\xc8\xa7"

    @staticmethod
    def printf_cmd(arg):
        """print with builtin printf"""
        return [SH, "-c", 'printf "$@"', "printf", arg]

    def test_subp_handles_bytestrings(self, tmp_path):
        """subp can run a bytestring command if shell is True."""
        tmp_file = str(tmp_path / "test.out")
        cmd = "echo HI MOM >> {tmp_file}".format(tmp_file=tmp_file)
        (out, _err) = subp.subp(cmd.encode("utf-8"), shell=True)
        assert "" == out
        assert "" == _err
        assert "HI MOM\n" == util.load_text_file(tmp_file)

    def test_subp_handles_strings(self, tmp_path):
        """subp can run a string command if shell is True."""
        tmp_file = str(tmp_path / "test.out")
        cmd = "echo HI MOM >> {tmp_file}".format(tmp_file=tmp_file)
        (out, _err) = subp.subp(cmd, shell=True)
        assert "" == out
        assert "" == _err
        assert "HI MOM\n" == util.load_text_file(tmp_file)

    def test_subp_handles_utf8(self):
        # The given bytes contain utf-8 accented characters as seen in e.g.
        # the "deja dup" package in Ubuntu.
        cmd = self.printf_cmd(self.utf8_valid_2)
        (out, _err) = subp.subp(cmd, capture=True)
        assert out == self.utf8_valid_2.decode("utf-8")

    def test_subp_respects_decode_false(self):
        (out, err) = subp.subp(
            self.stdin2out, capture=True, decode=False, data=self.utf8_valid
        )
        assert isinstance(out, bytes)
        assert isinstance(err, bytes)
        assert out == self.utf8_valid

    def test_subp_decode_ignore(self):
        """ensure that invalid utf-8 is ignored with the "ignore" kwarg"""
        # this executes a string that writes invalid utf-8 to stdout
        with mock.patch.object(
            subp.subprocess,
            "Popen",
            autospec=True,
        ) as sp:
            sp.return_value.communicate = mock.Mock(
                return_value=(b"abc\xaadef", None)
            )
            sp.return_value.returncode = 0
            assert (
                "abcdef"
                == subp.subp([SH], capture=True, decode="ignore").stdout
            )

    def test_subp_decode_strict_valid_utf8(self):
        (out, _err) = subp.subp(
            self.stdin2out, capture=True, decode="strict", data=self.utf8_valid
        )
        assert out == self.utf8_valid.decode("utf-8")

    def test_subp_decode_invalid_utf8_replaces(self):
        (out, _err) = subp.subp(
            self.stdin2out, capture=True, data=self.utf8_invalid
        )
        expected = self.utf8_invalid.decode("utf-8", "replace")
        assert out == expected

    def test_subp_decode_strict_raises(self):
        args = []
        kwargs = {
            "args": self.stdin2out,
            "capture": True,
            "decode": "strict",
            "data": self.utf8_invalid,
        }
        with pytest.raises(UnicodeDecodeError):
            subp.subp(*args, **kwargs)

    def test_subp_capture_stderr(self):
        data = b"hello world"
        (out, err) = subp.subp(
            self.stdin2err,
            capture=True,
            decode=False,
            data=data,
            update_env={"LC_ALL": "C"},
        )
        assert err == data
        assert out == b""

    def test_subp_reads_env(self):
        with mock.patch.dict("os.environ", values={"FOO": "BAR"}):
            assert {"FOO=BAR"}.issubset(
                subp.subp("env", capture=True).stdout.splitlines()
            )

    def test_subp_update_env(self):
        """test that subp's update_env argument updates the environment"""
        extra = {"FOO": "BAR", "HOME": "/root", "K1": "V1"}
        with mock.patch.dict("os.environ", values=extra):
            out, _err = subp.subp(
                "env",
                capture=True,
                update_env={"HOME": "/myhome", "K2": "V2"},
            )

        assert {"FOO=BAR", "HOME=/myhome", "K1=V1", "K2=V2"}.issubset(
            set(out.splitlines())
        )

    def test_subp_warn_missing_shebang(self, tmp_path):
        """Warn on no #! in script"""
        noshebang = str(tmp_path / "noshebang")
        util.write_file(noshebang, "true\n")

        print("os is %s" % os)
        os.chmod(noshebang, os.stat(noshebang).st_mode | stat.S_IEXEC)
        with pytest.raises(
            subp.ProcessExecutionError, match=r"Missing #! in script\?"
        ):
            subp.subp(
                (noshebang,),
            )

    def test_returns_none_if_no_capture(self):
        (out, err) = subp.subp(self.stdin2out, data=b"", capture=False)
        assert err is None
        assert out is None

    def test_exception_has_out_err_are_bytes_if_decode_false(self):
        """Raised exc should have stderr, stdout as bytes if no decode."""
        with pytest.raises(subp.ProcessExecutionError) as exc_info:
            subp.subp([BOGUS_COMMAND], decode=False)
        assert isinstance(exc_info.value.stdout, bytes)
        assert isinstance(exc_info.value.stderr, bytes)

    def test_exception_has_out_err_are_bytes_if_decode_true(self):
        """Raised exc should have stderr, stdout as string if no decode."""
        with pytest.raises(subp.ProcessExecutionError) as exc_info:
            subp.subp([BOGUS_COMMAND], decode=True)
        assert isinstance(exc_info.value.stdout, str)
        assert isinstance(exc_info.value.stderr, str)

    def test_exception_invalid_command(self):
        args = [None, "first", "arg", "missing"]
        with pytest.raises(subp.ProcessExecutionError):
            subp.subp(args)

    def test_bunch_of_slashes_in_path(self):
        assert "/target/my/path/" == subp.target_path("/target/", "//my/path/")
        assert "/target/my/path/" == subp.target_path(
            "/target/", "///my/path/"
        )

    def test_c_lang_can_take_utf8_args(self):
        """Independent of system LC_CTYPE, args can contain utf-8 strings.

        When python starts up, its default encoding gets set based on
        the value of LC_CTYPE.  If no system locale is set, the default
        encoding for both python2 and python3 in some paths will end up
        being ascii.

        Attempts to use setlocale or patching (or changing) os.environ
        in the current environment seem to not be effective.

        This test starts up a python with LC_CTYPE set to C so that
        the default encoding will be set to ascii.  In such an environment
        Popen(['command', 'non-ascii-arg']) would cause a UnicodeDecodeError.
        """
        python_prog = "\n".join(
            [
                "import json, sys",
                'sys.path.insert(0, "{}")'.format(get_top_level_dir()),
                "from cloudinit.subp import subp",
                "data = sys.stdin.read()",
                "cmd = json.loads(data)",
                "subp(cmd, capture=False)",
                "",
            ]
        )
        cmd = [
            SH,
            "-c",
            'printf "$@"',
            "--",
            self.utf8_valid.decode("utf-8"),
        ]
        python_subp = [sys.executable, "-c", python_prog]

        out, _err = subp.subp(
            python_subp,
            update_env={"LC_CTYPE": "C"},
            data=json.dumps(cmd).encode("utf-8"),
            decode=False,
        )
        assert self.utf8_valid == out



# This file is part of cloud-init. See LICENSE file for license information.

import os

from cloudinit.util import find_freebsd_part, get_path_dev_freebsd
from tests.unittests.helpers import get_distro, mock

M_PATH = "cloudinit.distros.freebsd."


class TestFreeBSD:
    @mock.patch(M_PATH + "subp.subp")
    def test_add_user(self, m_subp):
        distro = get_distro("freebsd")
        assert True is distro.add_user("me2", uid=1234, default=False)
        assert [
            mock.call(
                [
                    "pw",
                    "useradd",
                    "-n",
                    "me2",
                    "-u",
                    "1234",
                    "-d/home/me2",
                    "-m",
                ],
                logstring=[
                    "pw",
                    "useradd",
                    "-n",
                    "me2",
                    "-d/home/me2",
                    "-m",
                ],
            )
        ] == m_subp.call_args_list

    def test_unlock_passwd(self, caplog):
        distro = get_distro("freebsd")
        distro.unlock_passwd("me2")
        assert (
            "Dragonfly BSD/FreeBSD password lock is not reversible, "
            "ignoring unlock for user me2" in caplog.text
        )


class TestDeviceLookUp:
    @mock.patch("cloudinit.subp.subp")
    def test_find_freebsd_part_label(self, mock_subp):
        glabel_out = """
gptid/fa52d426-c337-11e6-8911-00155d4c5e47  N/A  da0p1
                              label/rootfs  N/A  da0p2
                                label/swap  N/A  da0p3
"""
        mock_subp.return_value = (glabel_out, "")
        res = find_freebsd_part("/dev/label/rootfs")
        assert "da0p2" == res

    @mock.patch("cloudinit.subp.subp")
    def test_find_freebsd_part_gpt(self, mock_subp):
        glabel_out = """
                                gpt/bootfs  N/A  vtbd0p1
gptid/3f4cbe26-75da-11e8-a8f2-002590ec6166  N/A  vtbd0p1
                                gpt/swapfs  N/A  vtbd0p2
                                gpt/rootfs  N/A  vtbd0p3
                            iso9660/cidata  N/A  vtbd2
"""
        mock_subp.return_value = (glabel_out, "")
        res = find_freebsd_part("/dev/gpt/rootfs")
        assert "vtbd0p3" == res

    @mock.patch("cloudinit.subp.subp")
    def test_find_freebsd_part_gptid(self, mock_subp):
        glabel_out = """
                                gpt/bootfs  N/A  vtbd0p1
                                gpt/efiesp  N/A  vtbd0p2
                                gpt/swapfs  N/A  vtbd0p3
gptid/4cd084b4-7fb4-11ee-a7ba-002590ec5bf2  N/A  vtbd0p4
"""
        mock_subp.return_value = (glabel_out, "")
        res = find_freebsd_part(
            "/dev/gptid/4cd084b4-7fb4-11ee-a7ba-002590ec5bf2"
        )
        assert "vtbd0p4" == res

    @mock.patch("cloudinit.subp.subp")
    def test_find_freebsd_part_ufsid(self, mock_subp):
        glabel_out = """
                                gpt/bootfs  N/A  vtbd0p1
                                gpt/efiesp  N/A  vtbd0p2
                                gpt/swapfs  N/A  vtbd0p3
                    ufsid/654e0663786f5131  N/A  vtbd0p4
"""
        mock_subp.return_value = (glabel_out, "")
        res = find_freebsd_part("/dev/ufsid/654e0663786f5131")
        assert "vtbd0p4" == res

    def test_get_path_dev_freebsd_label(self):
        mnt_list = """
/dev/label/rootfs  /                ufs     rw              1 1
devfs              /dev             devfs   rw,multilabel   0 0
fdescfs            /dev/fd          fdescfs rw              0 0
/dev/da1s1         /mnt/resource    ufs     rw              2 2
"""
        with mock.patch.object(os.path, "exists", return_value=True):
            res = get_path_dev_freebsd("/etc", mnt_list)
            assert res is not None



# This file is part of cloud-init. See LICENSE file for license information.

from datetime import datetime, timezone
from unittest import mock

import pytest

from cloudinit import version
from cloudinit.sources.azure import errors, kvp


@pytest.fixture()
def fake_utcnow():
    timestamp = datetime.now(timezone.utc)
    with mock.patch.object(kvp, "datetime", autospec=True) as m:
        m.now.return_value = timestamp
        yield timestamp


@pytest.fixture
def fake_vm_id(mocker):
    vm_id = "foo"
    mocker.patch(
        "cloudinit.sources.azure.kvp.identity.query_vm_id", return_value=vm_id
    )
    yield vm_id


@pytest.fixture
def telemetry_reporter(tmp_path):
    kvp_file_path = tmp_path / "kvp_pool_file"
    kvp_file_path.write_bytes(b"")
    reporter = kvp.handlers.HyperVKvpReportingHandler(
        kvp_file_path=str(kvp_file_path)
    )

    kvp.instantiated_handler_registry.register_item("telemetry", reporter)
    yield reporter
    kvp.instantiated_handler_registry.unregister_item("telemetry")


class TestReportFailureToHost:
    def test_report_via_kvp(self, caplog, fake_vm_id, telemetry_reporter):
        error = errors.ReportableError(reason="test")
        encoded_report = error.as_encoded_report(vm_id="fake-vm-id")

        assert kvp.report_via_kvp(encoded_report) is True
        assert (
            "KVP handler not enabled, skipping host report." not in caplog.text
        )

        report = {
            "key": "PROVISIONING_REPORT",
            "value": encoded_report,
        }
        assert report in list(telemetry_reporter._iterate_kvps(0))

    def test_report_skipped_without_telemetry(self, caplog):
        assert kvp.report_via_kvp("test report") is False
        assert "KVP handler not enabled, skipping host report." in caplog.text


class TestReportSuccessToHost:
    def test_report_success_to_host(
        self, caplog, fake_utcnow, fake_vm_id, telemetry_reporter
    ):
        assert kvp.report_success_to_host(vm_id=fake_vm_id) is True
        assert (
            "KVP handler not enabled, skipping host report." not in caplog.text
        )

        report_value = errors.encode_report(
            [
                "result=success",
                f"agent=Cloud-Init/{version.version_string()}",
                f"timestamp={fake_utcnow.isoformat()}",
                f"vm_id={fake_vm_id}",
            ]
        )

        report = {
            "key": "PROVISIONING_REPORT",
            "value": report_value,
        }
        assert report in list(telemetry_reporter._iterate_kvps(0))

    def test_report_skipped_without_telemetry(self, caplog):
        assert kvp.report_success_to_host(vm_id="fake") is False
        assert "KVP handler not enabled, skipping host report." in caplog.text


class TestGetKvpHandler:
    def test_vm_id_set_by_get_kvp_handler(
        self, telemetry_reporter, fake_vm_id
    ):
        """Test get_kvp_handler sets vm_id on the handler."""
        handler = kvp.get_kvp_handler()
        assert handler is telemetry_reporter

        # get_kvp_handler() should have set vm_id via identity.query_vm_id()
        assert handler.vm_id == fake_vm_id

        with mock.patch(
            "cloudinit.sources.azure.kvp.identity.query_vm_id",
            side_effect=AssertionError("should not be called twice"),
        ):
            handler_again = kvp.get_kvp_handler()
            assert handler_again is not None
            assert handler_again.vm_id == fake_vm_id

    def test_handles_vm_id_lookup_failure(self, telemetry_reporter, mocker):
        """Test handler gracefully handles identity.query_vm_id failure."""
        mocker.patch(
            "cloudinit.sources.azure.kvp.identity.query_vm_id",
            side_effect=RuntimeError("boom"),
        )
        # Also mock DMI to fail so we get ZERO_GUID
        mocker.patch(
            "cloudinit.reporting.handlers.dmi.read_dmi_data",
            return_value=None,
        )

        handler = kvp.get_kvp_handler()
        assert handler is telemetry_reporter
        # Falls back to DMI, which returns None, so stays ZERO_GUID
        assert (
            handler.vm_id == kvp.handlers.HyperVKvpReportingHandler.ZERO_GUID
        )



# This file is part of cloud-init. See LICENSE file for license information.

from cloudinit.distros.parsers import hosts

BASE_ETC = """
# Example
127.0.0.1	localhost
192.168.1.10	foo.mydomain.org  foo
192.168.1.10 	bar.mydomain.org  bar
146.82.138.7	master.debian.org      master
209.237.226.90	www.opensource.org
"""
BASE_ETC = BASE_ETC.strip()


class TestHostsHelper:
    def test_parse(self):
        eh = hosts.HostsConf(BASE_ETC)
        assert eh.get_entry("127.0.0.1") == [["localhost"]]
        assert eh.get_entry("192.168.1.10") == [
            ["foo.mydomain.org", "foo"],
            ["bar.mydomain.org", "bar"],
        ]
        eh = str(eh)
        assert eh.startswith("# Example")

    def test_add(self):
        eh = hosts.HostsConf(BASE_ETC)
        eh.add_entry("127.0.0.0", "blah")
        assert eh.get_entry("127.0.0.0") == [["blah"]]
        eh.add_entry("127.0.0.3", "blah", "blah2", "blah3")
        assert eh.get_entry("127.0.0.3") == [["blah", "blah2", "blah3"]]

    def test_del(self):
        eh = hosts.HostsConf(BASE_ETC)
        eh.add_entry("127.0.0.0", "blah")
        assert eh.get_entry("127.0.0.0") == [["blah"]]

        eh.del_entries("127.0.0.0")
        assert eh.get_entry("127.0.0.0") == []



# This file is part of cloud-init. See LICENSE file for license information.

from tests.unittests.helpers import get_distro


class TestAOSC:
    def test_get_distro(self):
        distro = get_distro("aosc")
        assert distro.osfamily == "aosc"
